import base64
import datetime

from pygeofilter import ast, values
from pygeofilter.parsers.fes.base import is_temporal, parse_temporal
from pygeofilter.parsers.fes.parser import FES11Parser as FES11ParserBug
from pygeofilter.parsers.fes.util import (Element, ParseInput, handle,
                                          handle_namespace)
from pygeofilter.util import parse_datetime, parse_duration
from pygml.georss import NAMESPACE as NAMESPACE_GEORSS
from pygml.georss import parse_georss
from pygml.pre_v32 import NAMESPACE as NAMESPACE_PRE_32
from pygml.pre_v32 import NSMAP as NSMAP_PRE_32
from pygml.pre_v32 import parse_pre_v32
from pygml.v32 import NAMESPACE as NAMESPACE_32
from pygml.v32 import NSMAP as NSMAP_32
from pygml.v32 import parse_v32
from pygml.v33 import NAMESPACE as NAMESPACE_33_CE
from pygml.v33 import parse_v33_ce


class FES11Parser(FES11ParserBug):
    """Workaround for missing PropertyName handling in fes 11 parser"""

    @handle("PropertyName")
    def property_name(self, node):
        return ast.Attribute(node.text)

    @handle("Filter")
    def filter_(self, node: Element, predicate):
        return predicate

    @handle("And")
    def and_(self, node: Element, lhs, rhs):
        return ast.And(lhs, rhs)

    @handle("Or")
    def or_(self, node: Element, lhs, rhs):
        return ast.Or(lhs, rhs)

    @handle("Not")
    def not_(self, node: Element, lhs):
        return ast.Not(lhs)

    @handle("PropertyIsEqualTo")
    def property_is_equal_to(self, node: Element, lhs, rhs):
        return ast.Equal(lhs, rhs)

    @handle("PropertyIsNotEqualTo")
    def property_is_not_equal_to(self, node: Element, lhs, rhs):
        return ast.NotEqual(lhs, rhs)

    @handle("PropertyIsLessThan")
    def property_is_less_than(self, node: Element, lhs, rhs):
        return ast.LessThan(lhs, rhs)

    @handle("PropertyIsGreaterThan")
    def property_is_greater_than(self, node: Element, lhs, rhs):
        return ast.GreaterThan(lhs, rhs)

    @handle("PropertyIsLessThanOrEqualTo")
    def property_is_less_than_or_equal_to(self, node: Element, lhs, rhs):
        return ast.LessEqual(lhs, rhs)

    @handle("PropertyIsGreaterThanOrEqualTo")
    def property_is_greater_than_or_equal_to(self, node: Element, lhs, rhs):
        return ast.GreaterEqual(lhs, rhs)

    @handle("PropertyIsLike")
    def property_is_like(self, node: Element, lhs, rhs):
        return ast.Like(
            lhs,
            rhs,
            wildcard=node.attrib["wildCard"],
            singlechar=node.attrib["singleChar"],
            escapechar=node.attrib.get("escape", node.attrib["escapeChar"]),
            nocase=node.attrib.get("matchCase", "true") == "false",
            not_=False,
        )

    @handle("PropertyIsNull")
    def property_is_null(self, node: Element, lhs):
        return ast.IsNull(lhs, not_=False)

    @handle("PropertyIsBetween")
    def property_is_between(self, node: Element, lhs, low, high):
        return ast.Between(lhs, low, high, False)

    @handle("LowerBoundary", "UpperBoundary")
    def boundary(self, node: Element, expression):
        return expression

    @handle("Equals")
    def geometry_equals(self, node: Element, lhs, rhs):
        return ast.GeometryEquals(lhs, rhs)

    @handle("Disjoint")
    def geometry_disjoint(self, node: Element, lhs, rhs):
        return ast.GeometryDisjoint(lhs, rhs)

    @handle("Touches")
    def geometry_touches(self, node: Element, lhs, rhs):
        return ast.GeometryTouches(lhs, rhs)

    @handle("Within")
    def geometry_within(self, node: Element, lhs, rhs):
        return ast.GeometryWithin(lhs, rhs)

    @handle("Overlaps")
    def geometry_overlaps(self, node: Element, lhs, rhs):
        return ast.GeometryOverlaps(lhs, rhs)

    @handle("Crosses")
    def geometry_crosses(self, node: Element, lhs, rhs):
        return ast.GeometryCrosses(lhs, rhs)

    @handle("Intersects")
    def geometry_intersects(self, node: Element, lhs, rhs):
        return ast.GeometryIntersects(lhs, rhs)

    @handle("Contains")
    def geometry_contains(self, node: Element, lhs, rhs):
        return ast.GeometryContains(lhs, rhs)

    @handle("DWithin")
    def distance_within(self, node: Element, lhs, rhs, distance_and_units):
        distance, units = distance_and_units
        return ast.DistanceWithin(lhs, rhs, distance, units)

    @handle("Beyond")
    def distance_beyond(self, node: Element, lhs, rhs, distance_and_units):
        distance, units = distance_and_units
        return ast.DistanceBeyond(lhs, rhs, distance, units)

    @handle("Distance")
    def distance(self, node: Element):
        return (float(node.text), node.attrib["uom"])

    # @handle('BBOX')
    # def geometry_bbox(self, node: Element, lhs, rhs):
    #     # TODO: ast.BBox() seems incompatible
    #     pass

    @handle("ValueReference")
    def value_reference(self, node):
        return ast.Attribute(node.text)

    @handle("Literal")
    def literal(self, node):
        type_ = node.get("type", "").rpartition(":")[2]
        value = node.text
        if type_ == "boolean":
            return value.lower() == "true"
        elif type_ in (
            "byte",
            "int",
            "integer",
            "long",
            "negativeInteger",
            "nonNegativeInteger",
            "nonPositiveInteger",
            "positiveInteger",
            "short",
            "unsignedByte",
            "unsignedInt",
            "unsignedLong",
            "unsignedShort",
        ):
            return int(value)
        elif type_ in ("decimal", "double", "float"):
            return float(value)
        elif type_ == "base64Binary":
            return base64.b64decode(value)
        elif type_ == "hexBinary":
            return bytes.fromhex(value)
        elif type_ == "date":
            return datetime.date.fromisoformat(value)
        elif type_ == "dateTime":
            return parse_datetime(value)
        elif type_ == "duration":
            return parse_duration(value)

        # return to string
        return value

    @handle_namespace(NAMESPACE_PRE_32, False)
    def gml_pre_32(self, node: Element):
        if is_temporal(node):
            return parse_temporal(node, NSMAP_PRE_32)

        return values.Geometry(parse_pre_v32(node))

    @handle_namespace(NAMESPACE_32, False)
    def gml_32(self, node: Element):
        if is_temporal(node):
            return parse_temporal(node, NSMAP_32)

        return values.Geometry(parse_v32(node))

    @handle_namespace(NAMESPACE_33_CE, False)
    def gml_33_ce(self, node: Element):
        return values.Geometry(parse_v33_ce(node))

    @handle_namespace(NAMESPACE_GEORSS, False)
    def georss(self, node: Element):
        return values.Geometry(parse_georss(node))


def parse(input_: ParseInput) -> ast.Node:
    return FES11Parser().parse(input_)



from eulxml.xmlmap import StringField, XmlObject


class PostRequest(XmlObject):

    operation = StringField(xpath="local-name()")
    version = StringField(xpath="./@version")
    service_type = StringField(xpath="./@service")



from eulxml.xmlmap import (FloatField, IntegerField, NodeField, NodeListField,
                           SimpleBooleanField, StringField, StringListField)

from ows_lib.xml_mapper.capabilities.mixins import (OGCServiceTypeMixin,
                                                    ReferenceSystemMixin)
from ows_lib.xml_mapper.capabilities.wms.mixins import (LayerMixin,
                                                        TimeDimensionMixin,
                                                        WebMapServiceMixin)
from ows_lib.xml_mapper.mixins import CustomXmlObject
from ows_lib.xml_mapper.namespaces import WMS_1_3_0_NAMESPACE, XLINK_NAMESPACE


class WebMapServiceDefaultSettings(CustomXmlObject):
    ROOT_NS = WMS_1_3_0_NAMESPACE
    ROOT_NAMESPACES = {
        "wms": WMS_1_3_0_NAMESPACE,
        "xlink": XLINK_NAMESPACE
    }


class ServiceMetadataContact(WebMapServiceDefaultSettings):
    ROOT_NAME = "ContactInformation"

    name = StringField(
        xpath="./wms:ContactPersonPrimary/wms:ContactOrganization")
    person_name = StringField(
        xpath="./wms:ContactPersonPrimary/wms:ContactPerson")
    phone = StringField(xpath="./wms:ContactVoiceTelephone")
    facsimile = StringField(xpath="./wms:ContactFacsimileTelephone")
    email = StringField(xpath="./wms:ContactElectronicMailAddress")
    country = StringField(xpath="./wms:ContactAddress/wms:Country")
    postal_code = StringField(xpath="./wms:ContactAddress/wms:PostCode")
    city = StringField(xpath="./wms:ContactAddress/wms:City")
    state_or_province = StringField(
        xpath="./wms:ContactAddress/wms:StateOrProvince")
    address = StringField(xpath="./wms:ContactAddress/wms:Address")


class ServiceType(OGCServiceTypeMixin, WebMapServiceDefaultSettings):
    ROOT_NAME = "WMS_Capabilities/@version='1.3.0'"

    version = StringField(xpath="./@version", choices='1.3.0')
    _name = StringField(xpath="./wms:Service/wms:Name")


class TimeDimension(TimeDimensionMixin, WebMapServiceDefaultSettings):
    """ Time Dimension in ISO8601 format"""

    ROOT_NAME = "Dimension[@name='time']"

    name = StringField(xpath="./@name", choices="time")
    units = StringField(xpath="./@units", choices="ISO8601")

    _extent = StringField(xpath="./text()")


class ReferenceSystem(ReferenceSystemMixin, WebMapServiceDefaultSettings):
    ROOT_NAME = "CRS"

    _ref_system = StringField(xpath=".")


class MimeType(WebMapServiceDefaultSettings):
    ROOT_NAME = "Format"
    mime_type = StringField(xpath="./wms:Format")


class LegendUrl(WebMapServiceDefaultSettings):
    ROOT_NAME = "LegendUrl"

    legend_url = StringField(
        xpath="./wms:OnlineResource/@xlink:href")
    height = IntegerField(xpath="./@height")
    width = IntegerField(xpath="./@width")
    mime_type = NodeField(xpath="./wms:Format", node_class=MimeType)


class Style(WebMapServiceDefaultSettings):
    ROOT_NAME = "Style"

    name = StringField(xpath="./wms:Name")
    title = StringField(xpath="./wms:Title")
    legend_url = NodeField(xpath="./wms:LegendURL", node_class=LegendUrl)


class RemoteMetadata(WebMapServiceDefaultSettings):
    ROOT_NAME = "OnlineResource/@xlink:href"
    link = StringField(
        xpath="./@xlink:href")


class Layer(LayerMixin, WebMapServiceDefaultSettings):
    ROOT_NAME = "Layer"

    title = StringField(xpath="./wms:Title")
    abstract = StringField(xpath="./wms:Abstract")
    keywords = StringListField(xpath="./wms:KeywordList/wms:Keyword")

    scale_min = FloatField(xpath="./wms:MinScaleDenominator")
    scale_max = FloatField(xpath="./wms:MaxScaleDenominator")

    _bbox_min_x = FloatField(
        xpath="./wms:EX_GeographicBoundingBox/wms:westBoundLongitude")
    _bbox_max_x = FloatField(
        xpath="./wms:EX_GeographicBoundingBox/wms:eastBoundLongitude")
    _bbox_min_y = FloatField(
        xpath="./wms:EX_GeographicBoundingBox/wms:southBoundLatitude")
    _bbox_max_y = FloatField(
        xpath="./wms:EX_GeographicBoundingBox/wms:northBoundLatitude")

    reference_systems = NodeListField(
        xpath="./wms:CRS", node_class=ReferenceSystem)
    identifier = StringField(xpath="./wms:Name")
    styles = NodeListField(xpath="./wms:Style", node_class=Style)
    is_queryable = SimpleBooleanField(
        xpath="./@queryable", true=1, false=0)
    is_opaque = SimpleBooleanField(xpath="./@opaque", true=1, false=0)
    is_cascaded = SimpleBooleanField(
        xpath="./@cascaded", true=1, false=0)
    dimensions = NodeListField(
        xpath="./wms:Dimension[@name='time']", node_class=TimeDimension)
    parent = NodeField(xpath="../../wms:Layer", node_class="self")
    children = NodeListField(xpath="./wms:Layer", node_class="self")

    remote_metadata = NodeListField(
        xpath="./wms:MetadataURL/wms:OnlineResource",
        node_class=RemoteMetadata)


class WebMapService(WebMapServiceMixin, WebMapServiceDefaultSettings):
    ROOT_NAME = "WMS_Capabilities/@version='1.3.0'"
    XSD_SCHEMA = "https://schemas.opengis.net/wms/1.3.0/capabilities_1_3_0.xsd"

    title = StringField(xpath="./wms:Service/wms:Title")
    abstract = StringField(xpath="./wms:Service/wms:Abstract")
    fees = StringField(xpath="./wms:Service/wms:Fees")
    access_constraints = StringField(
        xpath="./wms:Service/wms:AccessConstraints")
    service_url = StringField(
        xpath="./wms:Service/wms:OnlineResource/@xlink:href")

    # ForeignKey
    service_contact = NodeField(xpath="./wms:Service/wms:ContactInformation",
                                node_class=ServiceMetadataContact)

    # ManyToManyField
    keywords = StringListField(
        xpath="./wms:Service/wms:KeywordList/wms:Keyword")

    service_type = NodeField(xpath=".", node_class=ServiceType)

    root_layer = NodeField(
        xpath="./wms:Capability/wms:Layer", node_class=Layer)

    _layers = NodeListField(
        xpath="./wms:Capability//wms:Layer", node_class=Layer)

    # cause the information of operation urls are stored as entity name inside the xpath, we need to parse every operation url seperate.
    # To simplify the access of operation_urls we write a custom getter and setter property for it.
    # With that technique the usage of this mapper is easier and matches the db model
    _get_capabilities_mime_types = StringListField(
        xpath="./wms:Capability/wms:Request/wms:GetCapabilities/wms:Format")
    _get_capabilities_get_url = StringField(
        xpath="./wms:Capability/wms:Request/wms:GetCapabilities/wms:DCPType/wms:HTTP/wms:Get/wms:OnlineResource/@xlink:href")
    _get_capabilities_post_url = StringField(
        xpath="./wms:Capability/wms:Request/wms:GetCapabilities/wms:DCPType/wms:HTTP/wms:Post/wms:OnlineResource/@xlink:href")

    _get_map_mime_types = StringListField(
        xpath="./wms:Capability/wms:Request/wms:GetMap/wms:Format")
    _get_map_get_url = StringField(
        xpath="./wms:Capability/wms:Request/wms:GetMap/wms:DCPType/wms:HTTP/wms:Get/wms:OnlineResource/@xlink:href")
    _get_map_post_url = StringField(
        xpath="./wms:Capability/wms:Request/wms:GetMap/wms:DCPType/wms:HTTP/wms:Post/wms:OnlineResource/@xlink:href")

    _get_feature_info_mime_types = StringListField(
        xpath="./wms:Capability/wms:Request/wms:GetFeatureInfo/wms:Format")
    _get_feature_info_get_url = StringField(
        xpath="./wms:Capability/wms:Request/wms:GetFeatureInfo/wms:DCPType/wms:HTTP/wms:Get/wms:OnlineResource/@xlink:href")
    _get_feature_info_post_url = StringField(
        xpath="./wms:Capability/wms:Request/wms:GetFeatureInfo/wms:DCPType/wms:HTTP/wms:Post/wms:OnlineResource/@xlink:href")

    _describe_layer_mime_types = StringListField(
        xpath="./wms:Capability/wms:Request/wms:DescribeLayer/wms:Format")
    _describe_layer_get_url = StringField(
        xpath="./wms:Capability/wms:Request/wms:DescribeLayer/wms:DCPType/wms:HTTP/wms:Get/wms:OnlineResource/@xlink:href")
    _describe_layer_post_url = StringField(
        xpath="./wms:Capability/wms:Request/wms:DescribeLayer/wms:DCPType/wms:HTTP/wms:Post/wms:OnlineResource/@xlink:href")

    _get_legend_graphic_mime_types = StringListField(
        xpath="./wms:Capability/wms:Request/wms:GetLegendGraphic/wms:Format")
    _get_legend_graphic_get_url = StringField(
        xpath="./wms:Capability/wms:Request/wms:GetLegendGraphic/wms:DCPType/wms:HTTP/wms:Get/wms:OnlineResource/@xlink:href")
    _get_legend_graphic_post_url = StringField(
        xpath="./wms:Capability/wms:Request/wms:GetLegendGraphic/wms:DCPType/wms:HTTP/wms:Post/wms:OnlineResource/@xlink:href")

    _get_styles_mime_types = StringListField(
        xpath="./wms:Capability/wms:Request/wms:GetStyles/wms:Format")
    _get_styles_get_url = StringField(
        xpath="./wms:Capability/wms:Request/wms:GetStyles/wms:DCPType/wms:HTTP/wms:Get/wms:OnlineResource/@xlink:href")
    _get_styles_post_url = StringField(
        xpath="./wms:Capability/wms:Request/wms:GetStyles/wms:DCPType/wms:HTTP/wms:Post/wms:OnlineResource/@xlink:href")



from axis_order_cache.utils import get_epsg_srid
from django.contrib.gis.geos import GEOSGeometry
from eulxml import xmlmap


class Gml(xmlmap.XmlObject):
    srs_name = xmlmap.StringField(xpath="./@srsName")

    @property
    def to_geometry(self) -> GEOSGeometry:
        geometry = GEOSGeometry.from_gml(self.serialize())
        if self.srs_name:
            _, srid = get_epsg_srid(self.srs_name)
        else:
            srid = 4326  # default srs
        return GEOSGeometry(geo_input=geometry.wkt, srid=srid)



from django.http import HttpResponse


class SemanticError(Exception):
    """Basic exception for errors raised by wrong semantic parsed content"""


class OGCServiceException(HttpResponse):
    status_code = 200
    message = None
    locator = None

    def __init__(self, ogc_request, message: str = None, locator: str = None, *args, **kwargs):
        if message:
            self.message = message
        if locator:
            self.locator = locator
        self.ogc_request = ogc_request
        super().__init__(content_type="application/xml",
                         content=self.get_exception_string(), *args, **kwargs)

    def __eq__(self, __value: object) -> bool:
        return self.status_code == __value.status_code and self.message == __value.message and self.locator == __value.locator

    def get_locator(self):
        return self.locator

    def get_locator_string(self):
        return f'locator="{self.get_locator()}"'

    def get_message(self):
        return self.message

    def get_exception_string(self):
        return \
            '<?xml version="1.0" encoding="UTF-8"?>'\
            f'<ServiceExceptionReport version="{self.ogc_request.service_version}" xmlns="http://www.opengis.net/ogc" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://www.opengis.net/ogc">'\
            f'<ServiceException code="{self.code}" {self.get_locator_string() if self.get_locator() else ""}>'\
            f'{self.get_message()}'\
            '</ServiceException>'\
            '</ServiceExceptionReport>'


class MissingParameterException(OGCServiceException):
    code = "MissingParameter"


class MissingRequestParameterException(MissingParameterException):
    locator = "request"
    message = "Could not determine request method from http request."


class MissingVersionParameterException(MissingParameterException):
    locator = "version"
    message = "Could not determine version for the requested service."


class MissingServiceParameterException(MissingParameterException):
    locator = "service"
    message = "Could not determine service for the requested service."


class MissingConstraintLanguageParameterException(MissingParameterException):
    locator = "CONSTRAINTLANGUAGE"
    message = "CONSTRAINTLANGUAGE parameter is missing."


class InvalidParameterValueException(OGCServiceException):
    code = "IvalidParameterValue"


class OperationNotSupportedException(OGCServiceException):
    code = "OperationNotSupported"
    message = "No such operation"

    def get_locator(self):
        query_parameters = {
            k.lower(): v for k, v in self.request.GET.items()}
        return query_parameters.get('request', None)

    def get_message(self):
        return f"No such operation: {self.get_locator()}"


class LayerNotDefined(OGCServiceException):
    code = "LayerNotDefined"
    message = "unknown layer"
    locator = "LAYERS"



import copy
from typing import List

from django.contrib.gis.geos import GEOSGeometry
from django.contrib.gis.geos import Polygon as GeosPolygon
from eulxml.xmlmap import (NodeField, NodeListField, StringField,
                           StringListField, XmlObject)

from ows_lib.xml_mapper.consts import FES_AND, FES_OR, FES_WITHIN
from ows_lib.xml_mapper.namespaces import (FES_2_0_NAMEPSACE,
                                           GML_3_2_2_NAMESPACE,
                                           WFS_2_0_0_NAMESPACE)


class PolygonFilter(XmlObject):
    ROOT_NS = GML_3_2_2_NAMESPACE
    ROOT_NAME = "Within"
    ROOT_NAMESPACES = {
        "gml": GML_3_2_2_NAMESPACE
    }

    _srs_name = StringField(xpath="./@srsName")
    _position_list = StringField(
        xpath="./gml:exterior/gml:LinearRing/gml:posList")

    def __init__(self, srid=None, coords=None, node=None, context=None, **kwargs):
        super().__init__(node, context, **kwargs)
        if srid:
            self.srs_name = srid
        if coords:
            self.position_list = coords

    @property
    def srs_name(self):
        return self._srs_name

    @srs_name.setter
    def srs_name(self, srid):
        self._srs_name = f"urn:x-ogc:def:crs:EPSG:{srid}"

    @property
    def position_list(self):
        return self._position_list

    @position_list.setter
    def position_list(self, coords):
        position_list = ""
        for x, y in coords[0]:
            position_list += f"{x} {y} "
        self._position_list = position_list[:-1]


class WithinCondition(XmlObject):
    ROOT_NS = FES_2_0_NAMEPSACE
    ROOT_NAME = "Within"
    ROOT_NAMESPACES = {
        "fes": FES_2_0_NAMEPSACE,
        "gml": GML_3_2_2_NAMESPACE
    }

    value_reference = StringField(xpath="./fes:ValueReference")
    polygon_filter = NodeField(xpath="./gml:Polygon", node_class=PolygonFilter)


class OrCondition(XmlObject):
    ROOT_NS = FES_2_0_NAMEPSACE
    ROOT_NAME = "Or"
    ROOT_NAMESPACES = {
        "fes": FES_2_0_NAMEPSACE,
    }

    within_conditions = NodeListField(
        xpath=FES_WITHIN, node_class=WithinCondition)
    or_conditions = NodeListField(xpath=FES_OR, node_class="self")
    # FIXME: and_conditions are also possible


class AndCondition(XmlObject):
    ROOT_NS = FES_2_0_NAMEPSACE
    ROOT_NAME = "And"
    ROOT_NAMESPACES = {
        "fes": FES_2_0_NAMEPSACE,
    }

    within_conditions = NodeListField(
        xpath=FES_WITHIN, node_class=WithinCondition)
    and_conditions = NodeListField(xpath=FES_AND, node_class="self")
    or_conditions = NodeListField(xpath=FES_OR, node_class=OrCondition)


class Filter(XmlObject):
    ROOT_NS = FES_2_0_NAMEPSACE
    ROOT_NAME = "Filter"
    XSD_SCHEMA = "http://schemas.opengis.net/filter/2.0/filter.xsd"
    ROOT_NAMESPACES = {
        "fes": FES_2_0_NAMEPSACE,
        "gml": GML_3_2_2_NAMESPACE
    }

    ressource_ids = StringListField(xpath="./fes:ResourceId/@rid")

    and_condition = NodeField(xpath=FES_AND, node_class=AndCondition)
    or_condition = NodeField(xpath=FES_OR, node_class=OrCondition)
    within_conditions = NodeListField(
        xpath=FES_WITHIN, node_class=WithinCondition)


class Query(XmlObject):
    ROOT_NS = WFS_2_0_0_NAMESPACE
    ROOT_NAME = "Query"
    XSD_SCHEMA = "http://schemas.opengis.net/wfs/2.0/wfs.xsd"
    ROOT_NAMESPACES = {
        "wfs": WFS_2_0_0_NAMESPACE,
        "fes": FES_2_0_NAMEPSACE
    }

    # single typenames are seperated by ,
    # example: <Query typeNames="ms:Erdbebenstation_Schutzbereich, ms:Erdbebenereignisse">
    _type_names = StringField(xpath="./@typeNames")

    property_names = StringListField(xpath="./wfs:PropertyName")
    filter = NodeField(xpath="./fes:Filter", node_class=Filter)

    @property
    def type_names(self) -> List[str]:
        return self._type_names.split(",")

    @type_names.setter
    def type_names(self, type_names: List[str]) -> None:
        self._type_names = ", ".join(type_names)


class GetFeatureRequest(XmlObject):
    ROOT_NS = WFS_2_0_0_NAMESPACE
    ROOT_NAME = "GetFeature"
    XSD_SCHEMA = "http://schemas.opengis.net/wfs/2.0/wfs.xsd"
    ROOT_NAMESPACES = {
        "wfs": WFS_2_0_0_NAMESPACE,
        "fes": FES_2_0_NAMEPSACE,
        "gml": GML_3_2_2_NAMESPACE
    }

    # is optional, but mapserver for example will report an server error if you dont pass it.
    # Default is application/gml+xml; version=3.2
    output_format = StringField(xpath="./@outputFormat")
    version = StringField(xpath="./@version")
    service_type = StringField(xpath="./@service")

    # wfs 1.1.0 spec 9.2:
    # "The <GetFeature> element contains one or more <Query> elements, each of which in turn contains the description of a query." 
    queries = NodeListField(xpath="./wfs:Query", node_class=Query)

    def _construct_within_condition(self, srid: str, value_reference: str, coords) -> WithinCondition:
        within_condition = WithinCondition()
        within_condition.value_reference = value_reference
        within_condition.polygon_filter = PolygonFilter(
            srid=srid, coords=coords)
        return within_condition

    def _construct_spatial_filter_condition(self, polygon: GeosPolygon, value_reference: str):
        if len(polygon.coords) == 1:
            return self._construct_within_condition(srid=polygon.srid, value_reference=value_reference, coords=polygon.coords)
        elif len(polygon.coords) > 1:
            or_condition = OrCondition()
            for coords in polygon.coords:
                or_condition.within_conditions.append(self._construct_within_condition(
                    srid=polygon.srid, value_reference=value_reference, coords=coords))
            return or_condition

    def _append_spatial_filter_condition(self, polygon: GeosPolygon, value_reference: str, query: Query):
        filter_condition = self._construct_spatial_filter_condition(
            polygon=polygon, value_reference=value_reference)

        xml_node = query.filter.and_condition if query.filter.and_condition else query.filter

        if isinstance(filter_condition, WithinCondition):
            xml_node.within_conditions.append(filter_condition)
        elif isinstance(filter_condition, OrCondition):
            xml_node.or_condition = filter_condition

    def _adjust_filter_node(self, query: Query) -> None:
        if not query.filter:
            query.filter = Filter()
        elif not query.filter.and_condition:
            # Sourround the old filter with a fes:And node first to combine them binary together and secure the request spatial
            old_filter = copy.deepcopy(query.filter.node)
            query.filter = Filter()
            query.filter.and_condition = AndCondition()
            query.filter.and_condition.node.extend(
                [child for child in old_filter])

    def secure_spatial(self, feature_types: List[dict]) -> None:

        lookup_dict = {}
        for feature_type in feature_types:
            try:
                lookup_dict.update({
                    feature_type.get("type_name"): {
                        "geometry_property_name": feature_type.get("geometry_property_name"),
                        "allowed_area_union": feature_type.get("allowed_area_union")
                    }})
            except AttributeError:
                raise AttributeError(
                    "feature_types list shall be provides as a list of dicts with kind {'type_name': val, 'geometry_property_name': val, 'allowed_area_union': val}")

        query: Query
        for query in self.queries:
            _feature_type = lookup_dict.get(query.type_names[0], {})
            _polygon = _feature_type.get(
                "allowed_area_union", GEOSGeometry("POLYGON EMPTY"))
            _geometry_property_name = _feature_type.get(
                "geometry_property_name", "THE_GEOM")

            if len(query.type_names) == 1 and _polygon and not _polygon.empty:
                self._adjust_filter_node(query=query)
                self._append_spatial_filter_condition(
                    polygon=_polygon, value_reference=_geometry_property_name, query=query)

            elif len(query.type_names) > 1:
                raise NotImplementedError(
                    "Currently we can't secure a query with multple type names in a single query node.")

    @property
    def requested_feature_types(self) -> List[str]:
        """Collect all typeNames of the request."""
        type_names = []
        for query in self.queries:
            for type_name in query.type_names:
                type_names.append(type_name)
        return type_names



from typing import Callable, Dict, Iterable, List
from urllib import parse

from camel_converter import to_snake

from ows_lib.xml_mapper.exceptions import SemanticError
from ows_lib.xml_mapper.mixins import CallbackList


class OperationUrl:
    """Helper class to flatten operationurl information which are stored in the xml path."""

    def __init__(self, method: str, operation: str, mime_types: List[str], url: str, callback: Callable = None) -> None:
        self.__method = method
        self.__operation = operation
        self.__mime_types = mime_types
        self.__url = url
        self._callback = callback

    def __str__(self) -> str:
        return f"op: {self.operation} | http method: {self.method} | url: {self.url}"

    @property
    def method(self) -> str:
        return self.__method

    @method.setter
    def method(self, value: str) -> None:
        """Custom setter to trigger callback function to update xml nodes"""
        self.__method = value
        if self._callback:
            self._callback(self)

    @property
    def operation(self) -> str:
        return self.__operation

    @operation.setter
    def operation(self, value: str) -> None:
        self.__operation = value
        if self._callback:
            self._callback(self)

    @property
    def url(self) -> str:
        return self.__url

    @url.setter
    def url(self, value: str) -> None:
        self.__url = value
        if self._callback:
            self._callback(self)

    @property
    def mime_types(self) -> List[str]:
        return self.__mime_types

    @mime_types.setter
    def mime_types(self, value: List[str]):
        self.__mime_types = value
        if self._callback:
            self._callback(self)

    def transform_to_model(self) -> Dict:
        attr = {}
        if self.operation:
            attr.update({"operation": self.operation})
        if self.url:
            attr.update({"url": self.url})
        if self.method:
            attr.update({"method": self.method})
        return attr


class ReferenceSystemMixin:

    _code = None
    _prefix = None

    def __str__(self) -> str:
        return self._ref_system

    @property
    def _ref_system(self):
        raise NotImplementedError

    def _extract_ref_system(self) -> None:
        if "::" in self._ref_system:
            # example: ref_system = urn:ogc:def:crs:EPSG::4326
            code = self._ref_system.rsplit(":")[-1]
            prefix = self._ref_system.rsplit(":")[-3]
        elif ":" in self._ref_system:
            # example: ref_system = EPSG:4326
            code = self._ref_system.rsplit(":")[-1]
            prefix = self._ref_system.rsplit(":")[-2]
        else:
            raise SemanticError("reference system unknown")

        self._code = code
        self._prefix = prefix

    @property
    def code(self) -> str:
        if not self._code:
            self._extract_ref_system()
        return self._code

    @code.setter
    def code(self, new_code) -> None:
        self._ref_system = f"{self._prefix}:{new_code}"

    @property
    def prefix(self) -> str:
        if not self._prefix:
            self._extract_ref_system()
        return self._prefix

    @prefix.setter
    def prefix(self, new_prefix) -> None:
        self._ref_system = f"{new_prefix}:{self._code}"

    def transform_to_model(self) -> Dict:
        attr = super().transform_to_model()
        if self.prefix:
            attr.update({"prefix": self.prefix})
        if self.code:
            attr.update({"code": self.code})
        return attr


class OGCServiceTypeMixin:

    @property
    def _name(self):
        raise NotImplementedError

    @property
    def name(self) -> str:
        """ Custom property, cause the parsed name of the root element doesn't contains the right
            value for database. We need to parse again cause the root attribute contains different service type names
            as we store in our database.

            Returns:
                field_dict (dict): the dict which contains all needed information

            Raises:
                SemanticError if service name or version can not be found
        """
        if ":" in self._name:
            name = self._name.split(":", 1)[-1]
        elif " " in self._name:
            name = self._name.split(" ", 1)[-1]
        else:
            name = self._name

        name = name.lower()

        if name not in ["wms", "wfs", "csw"]:
            raise SemanticError(f"could not determine the service type for the parsed capabilities document. "
                                f"Parsed name was {name}")

        return name

    @name.setter
    def name(self, value: str) -> None:
        self._name = value


class OGCServiceMixin:
    """Abstract class for all OGCService xml mappers,
    which implements functionality for global usage."""
    _operation_urls: CallbackList = CallbackList([], None)

    @property
    def _possible_operations(self):
        raise NotImplementedError

    @property
    def operation_urls(self) -> List[OperationUrl]:
        """Custom getter to merge all operation urls as plane OperationUrl object."""
        if not self._operation_urls:
            _operation_urls = []

            for possible_operation in self._possible_operations:
                mime_types_attr_name = f"_{to_snake(possible_operation)}_mime_types"
                mime_types = getattr(self, mime_types_attr_name)

                get_url_attr_name = f"_{to_snake(possible_operation)}_get_url"
                get_url = getattr(self, get_url_attr_name)

                post_url_attr_name = f"_{to_snake(possible_operation)}_post_url"
                post_url = getattr(self, post_url_attr_name)

                if get_url:
                    _operation_urls.append(
                        OperationUrl(
                            method="Get",
                            operation=possible_operation,
                            mime_types=mime_types,
                            url=get_url,
                            callback=self._update_operation_url_xml_node)
                    )
                if post_url:
                    _operation_urls.append(
                        OperationUrl(
                            method="Post",
                            operation=possible_operation,
                            mime_types=mime_types,
                            url=post_url,
                            callback=self._update_operation_url_xml_node)
                    )

            self._operation_urls = CallbackList(
                _operation_urls, callback=self._handle_operation_urls_list_operation)

        return self._operation_urls

    @operation_urls.setter
    def operation_urls(self, operation_urls: List[OperationUrl]):
        self._clear_all_operation_urls()
        for operation_url in operation_urls:
            self._update_operation_url_xml_node_by_name(
                operation_url=operation_url)

    def _update_operation_url_xml_node_by_name(self, operation_url: OperationUrl, remove: bool = False):
        # find url attribute and update it
        url_attr_name = f"_{to_snake(operation_url.operation)}_{operation_url.method.lower()}_url"
        setattr(self, url_attr_name, None if remove else operation_url.url)

        # find mime_types list attribute and update it
        mime_types_attr_name = f"_{to_snake(operation_url.operation)}_mime_types"
        if remove and not self._operation_has_get_and_post(operation_url):
            setattr(self, mime_types_attr_name, [])
        elif not remove:
            mime_types = getattr(self, mime_types_attr_name)
            new_mime_types = filter(
                lambda mime_type: mime_type not in getattr(self, mime_types_attr_name), operation_url.mime_types)
            mime_types.extend(new_mime_types)

    def _update_operation_url_xml_node(self, operation_url: OperationUrl, remove: bool = False) -> None:
        if operation_url.operation in self._possible_operations:
            self._update_operation_url_xml_node_by_name(
                operation_url, remove)
            return
        raise ValueError(
            f"unsuported operation: {operation_url.operation}")

    def _handle_operation_urls_list_operation(self, list_operation, items: OperationUrl = None) -> None:
        """Custom setter to set/append new operation urls. The XML will be build implicitly by using this setter."""

        if isinstance(items, Iterable):
            for item in items:
                if not item._callback:
                    item._callback = self._update_operation_url_xml_node
        else:
            if items and not items._callback:
                items._callback = self._update_operation_url_xml_node

        if list_operation in ["append", "insert"]:
            self._update_operation_url_xml_node(items)
        elif list_operation == "extend":
            [self._update_operation_url_xml_node(item) for item in items]
        elif list_operation in ["pop", "remove"]:
            self._update_operation_url_xml_node(items, True)
        elif list_operation == "clear":
            self._clear_all_operation_urls()

    def _clear_all_operation_urls(self) -> None:
        for possible_operation in self._possible_operations:
            get_url_attr_name = f"_{to_snake(possible_operation)}_get_url"
            setattr(self, get_url_attr_name, None)

            post_url_attr_name = f"_{to_snake(possible_operation)}_post_url"
            setattr(self, post_url_attr_name, None)

            mime_types_attr_name = f"_{to_snake(possible_operation)}_mime_types"
            setattr(self, mime_types_attr_name, [])
        self._operation_urls = []

    def _operation_has_get_and_post(self, operation_url: OperationUrl) -> bool:
        return self.get_operation_url_by_name_and_method(name=operation_url.operation, method="Get") and self.get_operation_url_by_name_and_method(name=operation_url.operation, method="Post")

    def get_operation_url_by_name_and_method(self, name, method) -> OperationUrl:
        return next((operation_url for operation_url in self.operation_urls if operation_url.operation == name and operation_url.method == method), None)

    def camouflage_urls(self, new_domain: str, scheme: str = None) -> None:
        for operation_url in self.operation_urls:
            parsed = parse.urlparse(operation_url.url)
            if scheme:
                replaced: parse.ParseResult = parsed._replace(
                    netloc=new_domain, scheme=scheme)
            else:
                replaced: parse.ParseResult = parsed._replace(
                    netloc=new_domain)
            operation_url.url = replaced.geturl()

    def transform_to_model(self) -> Dict:
        attr = super().transform_to_model()
        if self.service_type and self.service_type.version:
            attr.update({"version": self.service_type.version})
        return attr



from django.db.models.query_utils import Q
from eulxml.xmlmap import NodeField, StringField, XmlObject
from lxml.etree import XMLSyntaxError
from pygeofilter.backends.django.evaluate import to_filter
from pygeofilter.parsers.ecql import parse as parse_cql
from pygeofilter.parsers.fes.util import NodeParsingError

from ows_lib.contrib.fes.parser import parse as parse_fes
from ows_lib.xml_mapper.exceptions import InvalidParameterValueException
from ows_lib.xml_mapper.namespaces import CSW_2_0_2_NAMESPACE, OGC_NAMESPACE


class GetRecordsRequest(XmlObject):
    ROOT_NS = CSW_2_0_2_NAMESPACE
    ROOT_NAME = "GetRecords"
    ROOT_NAMESPACES = {
        "csw": CSW_2_0_2_NAMESPACE,
        "ogc": OGC_NAMESPACE
    }

    service_version = StringField(xpath="./@version")
    service_type = StringField(xpath="./@service")
    result_type = StringField(xpath="./@resultType")
    start_position = StringField(xpath="./@startPosition")
    max_records = StringField(xpath="./@maxRecords")
    element_set_name = StringField(xpath="./csw:Query/csw:ElementSetName")
    type_names = StringField(xpath="./csw:Query/@typeNames")

    _sort_by_property = StringField(
        xpath='./csw:Query/ogc:SortBy/ogc:SortProperty/ogc:PropertyName')
    _sort_by_order = StringField(
        xpath='./csw:Query/ogc:SortBy/ogc:SortProperty/ogc:SortOrder')

    constraint_version = StringField(
        xpath='./csw:Query/csw:Constraint/@version')

    fes_filter = NodeField(
        xpath="./csw:Query/csw:Constraint/ogc:Filter", node_class=XmlObject)

    cql_filter = StringField(
        xpath="./csw:Query/csw:Constraint/csw:CqlText")

    @property
    def sort_by(self):
        prop = self._sort_by_property

        if self._sort_by_order == "DESC":
            return f"-{prop}"
        else:
            return prop

    @sort_by.setter
    def sort_by(self, value):
        if value is None:
            self._sort_by_order = None
            self._sort_by_property = None
        elif value.startswith("-"):
            self._sort_by_order = "DESC"
            self._sort_by_property = value.split("-")[1]
        else:
            self._sort_by_order = "ASC"
            self._sort_by_property = value

    def get_django_filter(self, field_mapping=None, mapping_choices=None):
        if not self.fes_filter and not self.cql_filter:
            return Q()
        elif self.fes_filter:
            try:
                ast = parse_fes(
                    self.fes_filter.node)
                return to_filter(ast, field_mapping, mapping_choices)
            except (XMLSyntaxError, NodeParsingError) as e:
                return InvalidParameterValueException(
                    ogc_request=self,
                    message=e
                )
        elif self.cql_filter:
            try:
                ast = parse_cql(
                    self.cql_filter.node
                )
                return to_filter(ast, field_mapping, mapping_choices)
            except Exception as e:
                return InvalidParameterValueException(
                    ogc_request=self,
                    message=e
                )



import urllib.parse
from typing import Dict, List

from axis_order_cache.registry import Registry
from axis_order_cache.utils import get_epsg_srid
from django.contrib.gis.gdal import SpatialReference
from django.contrib.gis.geos import GEOSGeometry, Polygon
from requests import Session

from ows_lib.client.exceptions import (MissingBboxParam, MissingCrsParam,
                                       MissingServiceParam)
from ows_lib.xml_mapper.capabilities.mixins import OGCServiceMixin


def update_queryparams(url: str, params: Dict):
    """Helper function to update query paramase inside an existing url with trailing query params"""
    url_parts = urllib.parse.urlparse(url)
    query = dict(urllib.parse.parse_qsl(url_parts.query))
    query.update(params)
    return url_parts._replace(query=urllib.parse.urlencode(query=query)).geturl()


def _construct_polygon_from_bbox_query_param_for_wfs(get_dict):
    """Construct a polygon from the parsed bbox query parameter, based on the given service type and version.

    **WFS 1.0.0 (see wfs specs - 6.2.8.2.3 BBOX)**:
    * Provides geographic coordinates in longitude/latitude | east/north and may not be trusted to
        respect the EPSG definition axis order. ==> mathematical x,y order is used.
    * The coordinate reference system of the bbox always matches the crs of the SRS query param.


    **WFS >1.0.0 (see wfs specs - 14.3.3 Bounding box)**:
    * Respects the axis order defined by the EPSG definition. ==> dynamic x,y order based on the result
        of the epsg registry is used.
    * The bbox values support n axis crs systems ==> BBOX=lcc1,lcc2,...,lccN,ucc1,ucc2,...uccN[,crsuri]
    * The coordinate reference system of the bbox by default is WGS84 (EPSG:4326). IF the bbox param provides a
    crsuri value this coordinate reference system shall be used instead.

    :return: the bbox parsed from the get_dict or an empty polygon if something went wrong.
    :rtype: :class:`django.contrib.gis.geos.polygon.Polygon`
    """
    try:
        major_version, minor_version, _ = get_dict["version"].split(
            ".")
        major_version = int(major_version)
        minor_version = int(minor_version)
        bbox = get_dict["bbox"]
        srid = get_dict.get("srs", None)
        if not srid:
            srid = get_dict.get("srsname", None)
            if not srid:
                raise MissingCrsParam

        xy_order = True

        if major_version == 1 and minor_version < 1:
            min_x, min_y, max_x, max_y = bbox.split(",")
            min_x = float(min_x)
            min_y = float(min_y)
            max_x = float(max_x)
            max_y = float(max_y)
        else:
            bbox_values = bbox.split(",")
            registry = Registry()
            if len(bbox_values) == 4:
                epsg_sr = registry.get(srid=4326)
            elif len(bbox_values) == 5:
                _, srid = get_epsg_srid(bbox_values[4])
                epsg_sr = registry.get(srid=srid)
            else:
                raise NotImplementedError(
                    "multiple dimension crs is not implemented.")
            min_x = float(bbox_values[0])
            min_y = float(bbox_values[1])
            max_x = float(bbox_values[2])
            max_y = float(bbox_values[3])
            xy_order = epsg_sr.is_xy_order

        if xy_order:
            return Polygon(((min_x, min_y), (min_x, max_y), (max_x, max_y), (max_x, min_y), (min_x, min_y)),
                           srid=srid)
        else:
            return Polygon(((min_y, min_x), (max_y, min_x), (max_y, max_x), (min_y, max_x), (min_y, min_x)),
                           srid=srid)
    except Exception:
        pass
    return GEOSGeometry('POLYGON EMPTY')


def _construct_polygon_from_bbox_query_param_for_wms(get_dict):
    """Construct a polygon from the parsed bbox query parameter, based on the given service type and version.

    * In WMS version < 1.3.0 requests with a Geographic Coordinate Reference System, the bbox is interpreted with
        X-axis ≙ Longitude and Y-axis ≙ Latitude value.
    * In WMS version >= 1.3.0 requests the bbox axis order is interpreted like the axis order of the requested
        reference system.

    .. note:: excerpt from ogc specs

        * **OGC WMS 1.1.0**: When the SRS parameter specifies a Geographic Coordinate Reference System, e.g.,
            "EPSG:4326", the returned image is implicitly projected using a pseudo-Plate Carrée projection that plots
            Longitude along the X-axis and Latitude along the Y-axis. The BBOX request parameter (Section 7.2.3.6)
            values for such a coordinate reference system shall be specified in the order minimum longitude, minimum
            latitude, maximum longitude, maximum latitude. The BBOX parameter values shall use the coordinate
            reference system units.
            Some Projected Coordinate Reference Systems, e.g., "EPSG:30800" ("RT38 2.5 gon W", used in Sweden), have
            axes order other than X=East, Y=North. The BBOX request parameter values for such a coordinate system
            shall be specified in the order minimum Easting, minimum Northing, maximum Easting, maximum Northing.
            The BBOX parameters shall use the coordinate reference system units. (see 6.5.5.1)
        * **OGC WMS 1.3.0**: EXAMPLE EPSG:4326 refers to WGS 84 geographic latitude, then longitude. That is, in
            this CRS the x axis corresponds to latitude, and the y axis to longitude. (see 6.7.3.3)

    :return: the bbox parsed from the get_dict or an empty polygon if something went wrong.
    :rtype: :class:`django.contrib.gis.geos.polygon.Polygon`
    """
    try:
        _, minor_version, _ = get_dict["version"].split(
            ".")
        minor_version = int(minor_version)
        srid = get_dict.get("srs", None)
        if not srid:
            srid = get_dict.get("crs", None)
            if not srid:
                raise MissingCrsParam

        bbox = get_dict["bbox"]
        min_x, min_y, max_x, max_y = bbox.split(",")
        min_x = float(min_x)
        min_y = float(min_y)
        max_x = float(max_x)
        max_y = float(max_y)

        xy_order = True
        if minor_version >= 3:
            sr = SpatialReference(srs_input=srid)
            registry = Registry()
            epsg_sr = registry.get(srid=sr.srid)
            srid = epsg_sr.srid
            xy_order = epsg_sr.is_xy_order
        else:
            if ":" in srid:
                srid = srid.split(":")[-1]
            srid = int(srid)

        if xy_order:
            return Polygon(((min_x, min_y), (min_x, max_y), (max_x, max_y), (max_x, min_y), (min_x, min_y)),
                           srid=srid)
        else:
            return Polygon(((min_y, min_x), (max_y, min_x), (max_y, max_x), (min_y, max_x), (min_y, min_x)),
                           srid=srid)
    except Exception:
        pass
    return GEOSGeometry('POLYGON EMPTY')


def construct_polygon_from_bbox_query_param(get_dict) -> GEOSGeometry:
    """Construct a polygon from the parsed bbox query parameter, based on the given service type and version.

    :return: the bbox parsed from the get_dict or an empty polygon if something went wrong.
    :rtype: :class:`django.contrib.gis.geos.geometry.GEOSGeometry`
    """
    service_type = get_dict.get("SERVICE", get_dict.get("service", None))
    if not service_type:
        raise MissingServiceParam
    if not get_dict.get("BBOX", get_dict.get("bbox", None)):
        raise MissingBboxParam
    if service_type.lower() == "wms":
        return _construct_polygon_from_bbox_query_param_for_wms(get_dict=get_dict)
    elif service_type.lower() == "wfs":
        return _construct_polygon_from_bbox_query_param_for_wfs(get_dict=get_dict)


def get_requested_layers(params: Dict) -> List[str]:
    """Filters the given params by requested layers

    :param params: all query parameters
    :type params: Dict
    :return: the requested layers from the query params
    :rtype: List[str]
    """
    return list(filter(None, params.get("LAYERS", params.get("layers", "")).split(",")))


def get_requested_feature_types(params: Dict) -> List[str]:
    """Filters the given params by requested featuretypes

    :param params: all query parameters
    :type params: Dict
    :return: the requested featuretypes from the query params
    :rtype: List[str]
    """
    return list(filter(None, params.get("TYPENAMES", params.get("typenames", "")).split(",")))


def get_requested_records(params: Dict) -> List[str]:
    """Filters the given params by requested featuretypes

    :param params: all query parameters
    :type params: Dict
    :return: the requested featuretypes from the query params
    :rtype: List[str]
    """
    return list(filter(None, params.get("Id", "").split(",")))


def get_client(capabilities: OGCServiceMixin, session: Session = Session()):
    """Helper function to construct the correct client version for given capabilities document

    :param capabilities: The parsed capabilities document
    :type capabilities: OGCServiceMixin
    :return: the concrete service client
    :rtype: WebMapService | WebFeatureService | CatalogueService
    """
    if capabilities.service_type.name == "wms":
        if capabilities.service_type.version == "1.1.1":
            from ows_lib.client.wms.wms111 import WebMapService
            return WebMapService(capabilities=capabilities, session=session)
    if capabilities.service_type.name == "wfs":
        if capabilities.service_type.version == "2.0.0":
            from ows_lib.client.wfs.wfs200 import WebFeatureService
            return WebFeatureService(capabilities=capabilities, session=session)
    if capabilities.service_type.name == "csw":
        if capabilities.service_type.version == "2.0.2":
            from ows_lib.client.csw.csw202 import CatalogueService
            return CatalogueService(capabilities=capabilities, session=session)




from eulxml.xmlmap import (IntegerField, NodeListField, SimpleBooleanField,
                           StringField, XmlObject)


class FeatureTypeProperty(XmlObject):
    ROOT_NAME = "element"

    max_occurs = IntegerField(xpath="./@maxOccurs")
    min_occurs = IntegerField(xpath="./@minOccurs")
    name = StringField(xpath="./@name")
    data_type = StringField(xpath="./@type")
    required = SimpleBooleanField(
        xpath="./@nillable", true="true", false="false")


class DescribeFeatureResponse(XmlObject):
    ROOT_NAME = "schema"
    XSD_SCHEMA = "https://www.w3.org/2001/XMLSchema.xsd"

    properties = NodeListField(
        xpath="./complexType/complexContent/extension/sequence/element", node_class=FeatureTypeProperty)



from typing import List

from django.contrib.gis.geos import Polygon
from ows_lib.xml_mapper.capabilities.mixins import (OGCServiceMixin,
                                                    ReferenceSystemMixin)


class FeatureTypeMixin:
    def __str__(self):
        return self.identifier

    @property
    def _bbox_lower_corner(self):
        raise NotImplementedError

    @property
    def _bbox_upper_corner(self):
        raise NotImplementedError

    @property
    def _default_crs_class(self):
        return NotImplementedError

    @property
    def _other_crs_class(self):
        return NotImplementedError

    @property
    def bbox_lat_lon(self) -> Polygon:
        # there is no default xmlmap field which parses to a geos polygon. So we convert it here.

        if self._bbox_lower_corner and self._bbox_upper_corner:
            min_x = float(self._bbox_lower_corner.split(" ")[0])
            min_y = float(self._bbox_lower_corner.split(" ")[1])
            max_x = float(self._bbox_upper_corner.split(" ")[0])
            max_y = float(self._bbox_upper_corner.split(" ")[1])

            return Polygon(
                ((min_x, min_y),
                 (min_x, max_y),
                 (max_x, max_y),
                 (max_x, min_y),
                 (min_x, min_y))
            )

    @bbox_lat_lon.setter
    def bbox_lat_lon(self, polygon: Polygon) -> None:
        # Custom setter function to mapp the geos Polygon object back to the xml attributes
        self._bbox_lower_corner = f"{polygon.extent[0]} {polygon.extent[1]}"
        self._bbox_upper_corner = f"{polygon.extent[2]} {polygon.extent[3]}"

    @property
    def reference_systems(self) -> List[ReferenceSystemMixin]:
        if self._default_reference_system and self._other_reference_systems:
            return [self._default_reference_system].extend(self._other_reference_systems)
        elif self._default_reference_system:
            return [self._default_reference_system]
        else:
            return []

    @reference_systems.setter
    def reference_systems(self, value: List[ReferenceSystemMixin]):
        """custom setter function to pass the first given reference system as default reference system"""
        first = True
        for crs in value:
            if first:
                self._default_reference_system = self._default_crs_class(
                    context={"_ref_system": crs.__str__()})
                first = False
            else:
                self._other_reference_systems = self._other_crs_class(
                    context={"_ref_system": crs.__str__()})


class WebFeatureServiceMixin(OGCServiceMixin):
    """Abstract class for WebMapService xml mappers,
    which implements functionality for global usage."""

    _possible_operations = ["GetCapabilities", "DescribeFeatureType",
                            "GetFeature", "GetPropertyValue", "ListStoredQueries", "DescribeStoredQueries", "CreateStoredQuery", "DropStoredQuery"]

    def get_feature_type_by_identifier(self, identifier: str):
        return next((feature_type for feature_type in self.feature_types if feature_type.identifier == identifier), None)



from importlib import import_module
from pathlib import Path

from eulxml.xmlmap import (StringField, XmlObject, load_xmlobject_from_file,
                           load_xmlobject_from_string)

from ows_lib.client.enums import OGCServiceEnum
from ows_lib.xml_mapper.exceptions import SemanticError
from ows_lib.xml_mapper.namespaces import NS_WC


def raise_default_sematic_error(kind: str):
    raise SemanticError(f"could not determine the service type for the parsed capabilities document. "
                        f"Parsed name was {kind}")


class OGCServiceTypeHelper(XmlObject):
    """Helper class to extract version and service type from capabilities document.
       Not for Common usage!
    """

    version = StringField(xpath="//@version")
    _kind = StringField(
        xpath=f"./{NS_WC}Service']/{NS_WC}Name']|./{NS_WC}ServiceIdentification']/{NS_WC}ServiceType']")

    @property
    def kind(self):
        if not self._kind:
            raise SemanticError(
                "could not determine the service type for the parsed capabilities document.")
        _kind = self._kind.lower()

        if ":" in _kind:
            _kind = _kind.split(":", 1)[-1]
        elif " " in _kind:
            _kind = _kind.split(" ", 1)[-1]

        if _kind not in ["wms", "wfs", "csw"]:
            raise_default_sematic_error(_kind)
        return _kind


def get_load_func(capabilities_xml):
    if isinstance(capabilities_xml, str) or isinstance(capabilities_xml, bytes):
        load_func = load_xmlobject_from_string
    elif isinstance(capabilities_xml, Path):
        capabilities_xml = capabilities_xml.resolve().__str__()
        load_func = load_xmlobject_from_file
    else:
        raise ValueError("xml must be ether a str or Path")
    return load_func


def get_ogc_service_type_helper(capabilities_xml: OGCServiceTypeHelper):
    load_func = get_load_func(capabilities_xml)
    return load_func(capabilities_xml, xmlclass=OGCServiceTypeHelper)


def get_import_path_for_xml_mapper(capabilities_xml) -> str:
    """helper function to resolve the correct import path for specific service type and version

    :param capabilities_xml: The capabilities raw file
    :type capabilities_xml: pathlib.Path | str | bytes

    :raises NotImplementedError: For all unsupported service types and versions

    :return: The relative path to the correct xml mapper
    :rtype: str

    """
    parsed_service: OGCServiceTypeHelper = get_ogc_service_type_helper(
        capabilities_xml=capabilities_xml)

    if parsed_service.kind == OGCServiceEnum.WMS.value:
        if parsed_service.version == "1.1.1":
            return "ows_lib.xml_mapper.capabilities.wms.wms111"
        elif parsed_service.version == "1.3.0":
            return "ows_lib.xml_mapper.capabilities.wms.wms130"
    elif parsed_service.kind == OGCServiceEnum.WFS.value:
        if parsed_service.version == "2.0.0":
            return "ows_lib.xml_mapper.capabilities.wfs.wfs200"
    elif parsed_service.kind == OGCServiceEnum.CSW.value:
        if parsed_service.version == "2.0.2":
            return "ows_lib.xml_mapper.capabilities.csw.csw202"

    raise NotImplementedError(
        f"Version {parsed_service.version} for {parsed_service.kind} is not supported.")


def get_xml_mapper(capabilities_xml):
    import_path = get_import_path_for_xml_mapper(
        capabilities_xml=capabilities_xml)

    parsed_service: OGCServiceTypeHelper = get_ogc_service_type_helper(
        capabilities_xml=capabilities_xml)

    if parsed_service.kind == OGCServiceEnum.WMS.value:
        return getattr(import_module(import_path), "WebMapService")
    elif parsed_service.kind == OGCServiceEnum.WFS.value:
        return getattr(import_module(import_path), "WebFeatureService")
    elif parsed_service.kind == OGCServiceEnum.CSW.value:
        return getattr(import_module(import_path), "CatalogueService")
    else:
        raise_default_sematic_error(parsed_service.kind)


def get_parsed_service(capabilities_xml):
    """Helper function to resolve the correct mapper class by the given capabilities document

    :param capabilities_xml: The capabilities raw file
    :type capabilities_xml: pathlib.Path | str | bytes
    :return: The concrete xml mapper class for specific service kind and version
    """
    load_func = get_load_func(capabilities_xml)
    xml_mapper = get_xml_mapper(capabilities_xml)
    return load_func(capabilities_xml, xmlclass=xml_mapper)



from typing import List

from requests import Request

from ows_lib.client.mixins import OgcClient
from ows_lib.client.utils import update_queryparams
from ows_lib.models.ogc_request import OGCRequest
from ows_lib.xml_mapper.xml_requests.wfs.get_feature import GetFeatureRequest


class WebFeatureServiceMixin(OgcClient):

    @property
    def type_name_qp(self):
        raise NotImplementedError

    @property
    def output_format_qp(self):
        raise NotImplementedError

    def describe_feature_type_request(
            self,
            type_names: List[str],
            output_format: List[str]) -> Request:
        params = {
            "VERSION": self.capabilitie.sservice_type.version,
            "REQUEST": "DescribeFeatureType",
            "SERVICE": self.capabilities.service_type.name,
            self.type_name_qp: ",".join(type_names),
            self.output_format_qp: ",".join(output_format)
        }
        url = update_queryparams(
            url=self.capabilities.get_operation_url_by_name_and_method(
                "DescribeFeatureType", "Get").url,
            params=params)
        return OGCRequest(method="GET", url=url)

    def get_feature_request(
            self,
            get_feature_request: GetFeatureRequest) -> OGCRequest:
        params = {
            "VERSION": self.capabilities.service_type.version,
            "REQUEST": "GetFeature",
            "SERVICE": self.capabilities.service_type.name,
        }
        url = update_queryparams(
            url=self.capabilities.get_operation_url_by_name_and_method(
                "GetFeature", "Post").url,
            params=params)

        return OGCRequest(method="POST", url=url, data=get_feature_request.serializeDocument(), headers={"Content-Type": "application/xml"})



import re
from typing import List

from ows_lib.client.mixins import OgcClient
from ows_lib.client.utils import update_queryparams
from ows_lib.models.ogc_request import OGCRequest


class CatalogueServiceMixin(OgcClient):

    def queryable_type_name(self):
        """Returns the first matching string of the constraints list which matches the name 'type'"""
        prog = re.compile(r'(\w+:type$)|(^type$)')
        if any((_match := prog.match(item)) for item in self.capabilities.get_records_constraints):
            return _match.group(0)
        else:
            return "type"

    def get_constraint(self, record_types: List[str]):
        type_name = self.queryable_type_name()
        record_types_filters = [
            f"<ogc:PropertyIsEqualTo><ogc:PropertyName>{type_name}</ogc:PropertyName><ogc:Literal>{record_type}</ogc:Literal></ogc:PropertyIsEqualTo>" for record_type in record_types]

        return f'<ogc:Filter xmlns:ogc="http://www.opengis.net/ogc">{"<ogc:Or>" if record_types_filters.__len__() > 1 else ""}{"".join(record_types_filters)}{"</ogc:Or>" if record_types_filters.__len__() > 1 else ""}</ogc:Filter>'

    def get_records_request(
        self,
        type_names: str = "gmd:MD_Metadata",
        result_type: str = "hits",
        output_schema: str = "http://www.isotc211.org/2005/gmd",
        element_set_name: str = "full",
        xml_constraint: str = None,
        max_records: int = None,
        start_position: int = None,

    ) -> OGCRequest:

        params = {
            "VERSION": self.capabilities.service_type.version,
            "REQUEST": "GetRecords",
            "SERVICE": "CSW",
            "typeNames": type_names,
            "resultType": result_type,
            "outputSchema": output_schema,
            "elementSetName": element_set_name,
        }

        if xml_constraint:
            params.update({
                "constraintLanguage": "FILTER",
                "CONSTRAINT_LANGUAGE_VERSION": "1.1.0",
                "Constraint": xml_constraint
            })

        if max_records:
            params.update({
                "maxRecords": max_records
            })

        if start_position:
            params.update({
                "startPosition": start_position
            })

        url = update_queryparams(
            url=self.capabilities.get_operation_url_by_name_and_method(
                "GetRecords", "Get").url,
            params=params)

        return OGCRequest(method="GET", url=url)

    def get_record_by_id_request(
        self,
        id: str,
        output_schema: str = "http://www.isotc211.org/2005/gmd",
        element_set_name: str = "full",
    ) -> OGCRequest:
        params = {
            "VERSION": self.capabilities.service_type.version,
            "SERVICE": "CSW",
            "REQUEST": "GetRecordById",
            "outputSchema": output_schema,
            "elementSetName": element_set_name,
            "id": id,
        }

        url = update_queryparams(
            url=self.capabilities.get_operation_url_by_name_and_method(
                "GetRecordById", "Get").url,
            params=params)

        return OGCRequest(method="GET", url=url)



from eulxml.xmlmap import StringField, StringListField, XmlObject

from ows_lib.xml_mapper.namespaces import CSW_2_0_2_NAMESPACE


class GetRecordByIdRequest(XmlObject):
    ROOT_NS = CSW_2_0_2_NAMESPACE
    ROOT_NAME = "GetRecordById"
    ROOT_NAMESPACES = {
        "csw": CSW_2_0_2_NAMESPACE,
    }

    service_version = StringField(xpath="./@version")
    service_type = StringField(xpath="./@service")
    element_set_name = StringField(xpath="./csw:ElementSetName")

    ids = StringListField(xpath="./csw:Id")



from eulxml.xmlmap import (DateTimeField, IntegerField, NodeListField,
                           StringField, XmlObject)

from ows_lib.xml_mapper.iso_metadata.iso_metadata import MdMetadata
from ows_lib.xml_mapper.namespaces import (CSW_2_0_2_NAMESPACE, GMD_NAMESPACE,
                                           XSI_NAMESPACE)


class CswRecord(XmlObject):
    ROOT_NAME = "Record"
    ROOT_NS = CSW_2_0_2_NAMESPACE
    ROOT_NAMESPACES = {
        "csw": CSW_2_0_2_NAMESPACE,
    }


class GetRecordsResponse(XmlObject):
    ROOT_NAME = "GetRecordsResponse"
    ROOT_NS = CSW_2_0_2_NAMESPACE
    ROOT_NAMESPACES = {
        "csw": CSW_2_0_2_NAMESPACE,
        "gmd": GMD_NAMESPACE,
        "xsi": XSI_NAMESPACE
    }

    XSD_SCHEMA = "http://schemas.opengis.net/csw/2.0.2/CSW-discovery.xsd"

    schema_location = StringField(
        xpath="./@xsi:schemaLocation"
    )

    # mandatory attributes
    total_records = IntegerField(
        xpath="./csw:SearchResults/@numberOfRecordsMatched")
    records_returned = IntegerField(
        xpath="./csw:SearchResults/@numberOfRecordsReturned")

    # optional attributes
    version = StringField(xpath="./@version")
    next_record = IntegerField(xpath="./csw:SearchResults/@nextRecord")
    record_schema = StringField(
        xpath="./csw:SearchResults/@recordSchema")
    element_set = StringField(
        xpath="./csw:SearchResults/@elementSet")
    result_set_id = StringField(
        xpath="./csw:SearchResults/@resultSetId")
    expires = DateTimeField(
        xpath="./csw:SearchResults/@expires")
    time_stamp = DateTimeField(
        xpath="./csw:SearchStatus/@timestamp")

    # record subelements
    csw_records = NodeListField(
        xpath="./csw:SearchResults/csw:Record", node_class=CswRecord)
    gmd_records = NodeListField(
        xpath="./csw:SearchResults/gmd:MD_Metadata", node_class=MdMetadata)

    def __init__(self, node=None, context=None, **kwargs):
        super().__init__(node, context, **kwargs)
        if not self.schema_location:
            self.schema_location = "http://www.opengis.net/cat/csw/2.0.2 http://schemas.opengis.net/csw/2.0.2/CSW-discovery.xsd"




from eulxml.xmlmap import IntegerField, NodeField, NodeListField, XmlObject
from ows_lib.xml_mapper.gml.gml import Gml
from ows_lib.xml_mapper.namespaces import (GML_3_2_2_NAMESPACE,
                                           WFS_2_0_0_NAMESPACE)


class Member(XmlObject):
    ROOT_NS = WFS_2_0_0_NAMESPACE
    ROOT_NAME = "member"
    ROOT_NAMESPACES = {
        "wfs": WFS_2_0_0_NAMESPACE
    }


class FeatureCollection(XmlObject):
    ROOT_NS = WFS_2_0_0_NAMESPACE
    ROOT_NAME = "FeatureCollection"
    ROOT_NAMESPACES = {
        "wfs": WFS_2_0_0_NAMESPACE,
        "gml": GML_3_2_2_NAMESPACE
    }

    number_matched: int = IntegerField(xpath="./@numberMatched")
    number_returned: int = IntegerField(xpath="./@numberReturned")

    members: Member = NodeListField(xpath="./wfs:member", node_class=Member)

    bounded_by: Gml = NodeField(xpath="./wfs:boundedBy/gml:*", node_class=Gml)



GEOMETRY_DATA_TYPES = [
    "gml:MultiPolygonPropertyType",
    "gml:MultiSurfacePropertyType",
    "gml:PolygonPropertyType",
    "gml:SurfacePropertyType",
    "gml:MultiLineStringPropertyType",
    "gml:MultiCurvePropertyType",
    "gml:LineStringPropertyType",
    "gml:GeometryPropertyType",
    "gml:CurvePropertyType",
    "gml:PointPropertyType",
    "gml:MultiPointPropertyType"
]



from eulxml.xmlmap import DateTimeField, NodeField, StringField, XmlObject

from ows_lib.xml_mapper.namespaces import CSW_2_0_2_NAMESPACE


class Acknowledgement(XmlObject):
    ROOT_NAME = "Acknowledgement"
    ROOT_NS = CSW_2_0_2_NAMESPACE
    ROOT_NAMESPACES = {
        "csw": CSW_2_0_2_NAMESPACE,

    }

    # mandatory attributes
    time_stamp = DateTimeField(
        xpath="./@timestamp")

    echoed_get_records_request = NodeField(
        xpath="./csw:EchoedRequest/csw:GetRecords", node_class=XmlObject
    )

    request_id = StringField(
        xpath="./csw:RequestId"
    )



from typing import Dict, List, Tuple

from requests import Request

from ows_lib.client.mixins import OgcClient
from ows_lib.client.utils import update_queryparams
from ows_lib.models.ogc_request import OGCRequest


class WebMapServiceMixin(OgcClient):
    """Abstract WebMapService class which implements some basic functionality for all wms client applications
    """
    @property
    def crs_qp(self):
        raise NotImplementedError

    @property
    def get_map_operation_name(self):
        raise NotImplementedError

    @property
    def get_feature_info_operation_name(self):
        raise NotImplementedError

    def get_map_request(
            self,
            layers: List[str],
            styles: List[str],
            crs: str,
            bbox: Tuple[float, float, float, float],
            width: int,
            height: int,
            format: str,
            transparent: bool = None,
            bgcolor: int = 0xFFFFFF,
            exceptions: str = "xml",
            time: List[str] = None,
            elevation: float = None) -> OGCRequest:
        """Constructs a GetMap request to use for requesting

        :param layers: The name of layers which shall be requested
        :type layers: List[str]
        :param styles: The styles of the layers which shall be requested
        :type styles: List[str]
        :param crs: the reference system which shall be used
        :type crs: str
        :param bbox: the bounding box 
        :type bbox: Tuple[float, float, float, float]
        :param width: the pixel width 
        :type width: int
        :param height: the pixel height
        :type height: int
        :param format: the response format
        :type format: str
        :param transparent: shall the response be a transparent image?, defaults to None
        :type transparent: bool, optional
        :param bgcolor: the background color of the response, defaults to 0xFFFFFF
        :type bgcolor: int, optional
        :param exceptions: the exception format which shall be used by the server, defaults to "xml"
        :type exceptions: str, optional
        :param time: the time value or range for map data, defaults to None
        :type time: List[str], optional
        :param elevation: _description_, defaults to None
        :type elevation: float, optional
        :return: the constructed get map request object
        :rtype: Request
        """

        if isinstance(transparent, str):
            if transparent == "TRUE":
                transparent = True
            else:
                transparent = False

        params = {
            "VERSION": self.capabilities.service_type.version,
            "REQUEST": "GetMap",
            "SERVICE": self.capabilities.service_type.name,
            "LAYERS": ",".join(layers),
            "STYLES": ",".join(styles),
            self.crs_qp: crs,
            "BBOX": ",".join(bbox),
            "WIDTH": width,
            "HEIGHT": height,
            "FORMAT": format,
            "TRANSPARENT": transparent,
            "BGCOLOR": hex(bgcolor),
            "EXCEPTIONS": exceptions,
        }
        if time:
            params.update({"TIME": ",".join(time)})
        if elevation:
            params.update({"ELEVATION": elevation})

        url = update_queryparams(
            url=self.capabilities.get_operation_url_by_name_and_method(
                self.get_map_operation_name, "Get").url,
            params=params)

        return Request(method="GET", url=url)

    def get_feature_info_request(
            self,
            get_map_request: Request,
            query_layers: List[str],
            info_format: str,
            i: int,
            j: int,
            feature_count: int = 0,
            exceptions: str = "xml") -> OGCRequest:
        """Constructs a GetFeatureInfo request to use for requesting

        :param get_map_request: The GetMap request where this request shall based on
        :type get_map_request: requests.Request
        :param query_layers: The list of layers for that the feature info shall be requested
        :type query_layers: List[str]
        :param info_format: The concrete format of the response
        :type info_format: str
        :param i: the x value of the x/y point tuple
        :type i: int
        :param j: the y value of the x/y point tuple
        :type j: int
        :param feature_count: The number of features that shall be returned, defaults to 0
        :type feature_count: int, optional
        :param exceptions: the exception format of the server to response with, defaults to "xml"
        :type exceptions: str, optional
        :return: the constructed get feature info request object
        :rtype: requests.Request
        """
        params: Dict = get_map_request.params

        get_feature_info_params = {
            "VERSION": self.capabilities.service_type.version,
            "REQUEST": "GetFeatureInfo",
            "SERVICE": self.capabilities.service_type.name,
            "QUERY_LAYERS": ",".join(query_layers),
            "INFO_FORMAT": info_format,
            "FEATURE_COUNT": feature_count,
            "I": i,
            "J": j,
            "EXCEPTIONS": exceptions,
        }

        params.update(get_feature_info_params)

        url = update_queryparams(
            url=self.capabilities.get_operation_url_by_name_and_method(
                self.get_feature_info_operation_name, "Get").url,
            params=params)

        return OGCRequest(method="GET", url=url)



from eulxml.xmlmap import (NodeField, NodeListField, StringField,
                           StringListField)

from ows_lib.xml_mapper.capabilities.mixins import (OGCServiceTypeMixin,
                                                    ReferenceSystemMixin)
from ows_lib.xml_mapper.capabilities.wfs.mixins import (FeatureTypeMixin,
                                                        WebFeatureServiceMixin)
from ows_lib.xml_mapper.mixins import CustomXmlObject
from ows_lib.xml_mapper.namespaces import (FES_2_0_NAMEPSACE,
                                           GML_3_2_2_NAMESPACE,
                                           OWS_1_1_NAMESPACE,
                                           WFS_2_0_0_NAMESPACE,
                                           XLINK_NAMESPACE)


class WebFeatureServiceDefaultSettings(CustomXmlObject):
    ROOT_NS = WFS_2_0_0_NAMESPACE
    ROOT_NAMESPACES = {
        "wms": WFS_2_0_0_NAMESPACE,
        "ows": OWS_1_1_NAMESPACE,
        "gml": GML_3_2_2_NAMESPACE,
        "fes": FES_2_0_NAMEPSACE,
        "xlink": XLINK_NAMESPACE
    }


class ServiceMetadataContact(WebFeatureServiceDefaultSettings):
    ROOT_NAME = "ows:ServiceProvider"

    name = StringField(xpath="./ows:ProviderName")
    person_name = StringField(xpath="./ows:ServiceContact/ows:IndividualName")

    phone = StringField(
        xpath="./ows:ServiceContact/ows:ContactInfo/ows:Phone/ows:Voice")
    facsimile = StringField(
        xpath="./ows:ServiceContact/ows:ContactInfo/ows:Phone/ows:Facsimile")
    email = StringField(
        xpath="./ows:ServiceContact/ows:ContactInfo/ows:Address/ows:ElectronicMailAddress")
    country = StringField(
        xpath="./ows:ServiceContact/ows:ContactInfo/ows:Address/ows:Country")
    postal_code = StringField(
        xpath="./ows:ServiceContact/ows:ContactInfo/ows:Address/ows:PostalCode")
    city = StringField(
        xpath="./ows:ServiceContact/ows:ContactInfo/ows:Address/ows:City")
    state_or_province = StringField(
        xpath="./ows:ServiceContact/ows:ContactInfo/ows:Address/ows:AdministrativeArea")
    address = StringField(
        xpath="./ows:ServiceContact/ows:ContactInfo/ows:Address/ows:DeliveryPoint")


class ServiceType(OGCServiceTypeMixin, WebFeatureServiceDefaultSettings):
    ROOT_NAME = "WFS_Capabilities/@version='2.0.0'"

    version = StringField(xpath="./@version", choices='2.0.0')
    _name = StringField(xpath="./ows:ServiceIdentification/ows:ServiceType")


class ReferenceSystem(ReferenceSystemMixin, WebFeatureServiceDefaultSettings):
    _ref_system = StringField(xpath=".")


class DefaultReferenceSystem(ReferenceSystem):
    ROOT_NAME = "DefaultCRS"


class OtherReferenceSystem(ReferenceSystem):
    ROOT_NAME = "OtherCRS"


class OutputFormat(WebFeatureServiceDefaultSettings):
    ROOT_NAME = "Format"

    mime_type = StringField(xpath="./wfs:Format")


class RemoteMetadata(WebFeatureServiceDefaultSettings):
    ROOT_NAME = "MetadataURL"

    link = StringField(xpath="./@xlink:href")


class FeatureType(FeatureTypeMixin, WebFeatureServiceDefaultSettings):
    ROOT_NAME = "FeatureType"

    identifier = StringField(xpath="./wfs:Name")
    title = StringField(xpath="wfs:Title")
    abstract = StringField(xpath="wfs:Abstract")

    remote_metadata = NodeListField(
        xpath="./wfs:MetadataURL", node_class=RemoteMetadata)

    _bbox_lower_corner = StringField(
        xpath="./ows:WGS84BoundingBox/ows:LowerCorner")
    _bbox_upper_corner = StringField(
        xpath="./ows:WGS84BoundingBox/ows:UpperCorner")

    output_formats = NodeListField(
        xpath="./wfs:OutputFormats/wfs:Format", node_class=OutputFormat)

    _default_crs_class = DefaultReferenceSystem
    _other_crs_class = OtherReferenceSystem
    _default_reference_system = NodeField(
        xpath="./wfs:DefaultCRS", node_class=_default_crs_class)
    _other_reference_systems = NodeListField(
        xpath="./wfs:OtherCRS", node_class=_other_crs_class)

    keywords = StringListField(xpath="./ows:Keywords/ows:Keyword")


class WebFeatureService(WebFeatureServiceMixin, WebFeatureServiceDefaultSettings):
    ROOT_NAME = "WFS_Capabilities"
    XSD_SCHEMA = "http://www.opengis.net/wfs/2.0"

    title = StringField(xpath="./ows:ServiceIdentification/ows:Title")
    abstract = StringField(xpath="./ows:ServiceIdentification/ows:Abstract")
    fees = StringField(xpath="./ows:ServiceIdentification/ows:Fees")
    access_constraints = StringField(
        xpath="./ows:ServiceIdentification/ows:AccessConstraints")

    # ForeignKey
    service_contact = NodeField(xpath="./ows:ServiceProvider",
                                node_class=ServiceMetadataContact)

    # ManyToManyField
    keywords = StringListField(
        xpath="./ows:ServiceIdentification/ows:Keywords/ows:Keyword")

    service_type = NodeField(xpath=".", node_class=ServiceType)

    feature_types = NodeListField(xpath="./wfs:FeatureTypeList/wfs:FeatureType",
                                  node_class=FeatureType)

    # cause the information of operation urls are stored as entity name inside the xpath, we need to parse every operation url seperate.
    # To simplify the access of operation_urls we write a custom getter and setter property for it.
    # With that technique the usage of this mapper is easier and matches the db model
    _get_capabilities_mime_types = StringListField(
        xpath="./ows:OperationsMetadata/ows:Operation[@name='GetCapabilities']/ows:Parameter[@name='AcceptFormats']/ows:AllowedValues/ows:Value")
    _get_capabilities_get_url = StringField(
        xpath="./ows:OperationsMetadata/ows:Operation[@name='GetCapabilities']/ows:DCP/ows:HTTP/ows:Get/@xlink:href")
    _get_capabilities_post_url = StringField(
        xpath="./ows:OperationsMetadata/ows:Operation[@name='GetCapabilities']/ows:DCP/ows:HTTP/ows:Post/@xlink:href")

    _describe_feature_type_mime_types = StringListField(
        xpath="./ows:OperationsMetadata/ows:Operation[@name='DescribeFeatureType']/ows:Parameter[@name='outputFormat']/ows:AllowedValues/ows:Value")
    _describe_feature_type_get_url = StringField(
        xpath="./ows:OperationsMetadata/ows:Operation[@name='DescribeFeatureType']/ows:DCP/ows:HTTP/ows:Get/@xlink:href")
    _describe_feature_type_post_url = StringField(
        xpath="./ows:OperationsMetadata/ows:Operation[@name='DescribeFeatureType']/ows:DCP/ows:HTTP/ows:Post/@xlink:href")

    _get_feature_mime_types = StringListField(
        xpath="./ows:OperationsMetadata/ows:Operation[@name='GetFeature']/ows:Parameter[@name='outputFormat']/ows:AllowedValues/ows:Value")
    _get_feature_get_url = StringField(
        xpath="./ows:OperationsMetadata/ows:Operation[@name='GetFeature']/ows:DCP/ows:HTTP/ows:Get/@xlink:href")
    _get_feature_post_url = StringField(
        xpath="./ows:OperationsMetadata/ows:Operation[@name='GetFeature']/ows:DCP/ows:HTTP/ows:Post/@xlink:href")

    _get_property_value_mime_types = StringListField(
        xpath="./ows:OperationsMetadata/ows:Operation[@name='GetPropertyValue']/ows:Parameter[@name='outputFormat']/ows:AllowedValues/ows:Value")
    _get_property_value_get_url = StringField(
        xpath="./ows:OperationsMetadata/ows:Operation[@name='GetPropertyValue']/ows:DCP/ows:HTTP/ows:Get/@xlink:href")
    _get_property_value_post_url = StringField(
        xpath="./ows:OperationsMetadata/ows:Operation[@name='GetPropertyValue']/ows:DCP/ows:HTTP/ows:Post/@xlink:href")

    _list_stored_queries_mime_types = StringListField(
        xpath="./ows:OperationsMetadata/ows:Operation[@name='ListStoredQueries']/ows:Parameter[@name='outputFormat']/ows:AllowedValues/ows:Value")
    _list_stored_queries_get_url = StringField(
        xpath="./ows:OperationsMetadata/ows:Operation[@name='ListStoredQueries']/ows:DCP/ows:HTTP/ows:Get/@xlink:href")
    _list_stored_queries_post_url = StringField(
        xpath="./ows:OperationsMetadata/ows:Operation[@name='ListStoredQueries']/ows:DCP/ows:HTTP/ows:Post/@xlink:href")

    _describe_stored_queries_mime_types = StringListField(
        xpath="./ows:OperationsMetadata/ows:Operation[@name='DescribeStoredQueries']/ows:Parameter[@name='outputFormat']/ows:AllowedValues/ows:Value")
    _describe_stored_queries_get_url = StringField(
        xpath="./ows:OperationsMetadata/ows:Operation[@name='DescribeStoredQueries']/ows:DCP/ows:HTTP/ows:Get/@xlink:href")
    _describe_stored_queries_post_url = StringField(
        xpath="./ows:OperationsMetadata/ows:Operation[@name='DescribeStoredQueries']/ows:DCP/ows:HTTP/ows:Post/@xlink:href")

    _create_stored_query_mime_types = StringListField(
        xpath="./ows:OperationsMetadata/ows:Operation[@name='CreateStoredQuery']/ows:Parameter[@name='outputFormat']/ows:AllowedValues/ows:Value")
    _create_stored_query_get_url = StringField(
        xpath="./ows:OperationsMetadata/ows:Operation[@name='CreateStoredQuery']/ows:DCP/ows:HTTP/ows:Get/@xlink:href")
    _create_stored_query_post_url = StringField(
        xpath="./ows:OperationsMetadata/ows:Operation[@name='CreateStoredQuery']/ows:DCP/ows:HTTP/ows:Post/@xlink:href")

    _drop_stored_query_mime_types = StringListField(
        xpath="./ows:OperationsMetadata/ows:Operation[@name='DropStoredQuery']/ows:Parameter[@name='outputFormat']/ows:AllowedValues/ows:Value")
    _drop_stored_query_get_url = StringField(
        xpath="./ows:OperationsMetadata/ows:Operation[@name='DropStoredQuery']/ows:DCP/ows:HTTP/ows:Get/@xlink:href")
    _drop_stored_query_post_url = StringField(
        xpath="./ows:OperationsMetadata/ows:Operation[@name='DropStoredQuery']/ows:DCP/ows:HTTP/ows:Post/@xlink:href")

    # TODO: more Operations possible: LockFeature, GetFeatureWithLock, Transaction



from typing import Union

from lxml import etree
from pygeofilter import ast
from pygeofilter.parsers.fes.util import Element, ElementTree
from pygeofilter.parsers.fes.v20 import FES20Parser

from .v11.parser import FES11Parser


def parse(xml: Union[str, Element, ElementTree]) -> ast.Node:
    """Workaround for missing PropertyName handling in fes 11 parser"""

    if isinstance(xml, str):
        root = etree.fromstring(xml)
    else:
        root = xml

    # decide upon namespace which parser to use
    namespace = etree.QName(root).namespace
    if namespace == FES11Parser.namespace:
        return FES11Parser().parse(root)
    elif namespace == FES20Parser.namespace:
        return FES20Parser().parse(root)

    raise ValueError(f"Unsupported namespace {namespace}")



from abc import ABC

from requests import Request, Response, Session

from ows_lib.client.exceptions import InitialError
from ows_lib.client.utils import update_queryparams
from ows_lib.models.ogc_request import OGCRequest
from ows_lib.xml_mapper.capabilities.mixins import OGCServiceMixin
from ows_lib.xml_mapper.utils import get_parsed_service


class OgcClient(ABC):
    """Abstract OgcClient class which implements some basic functionality for all ogc client applications

    :param capabilities: The capabilities document to initialize the client
    :type capabilities: OGCServiceMixin | str

    :param session: The session object that shall be used
    :type session: requests.Session, optional
    """
    capabilities: OGCServiceMixin

    def __init__(
            self,
            capabilities,
            session: Session = Session(),
            *args,
            **kwargs):
        super().__init__(*args, **kwargs)

        self.session = session

        if isinstance(capabilities, OGCServiceMixin):
            self.capabilities = capabilities
        elif capabilities is str and "?" in capabilities:
            # client was initialized with an url
            response = self.send_request(
                request=Request(method="GET", url=capabilities))
            if response.status_code <= 202 and "xml" in response.headers["content-type"]:
                self.capabilities = get_parsed_service(response.content)
            else:
                raise InitialError(
                    f"client could not be initialized by the given url: {capabilities}. Response status code: {response.status_code}")

    def send_request(self, request: OGCRequest, timeout: int = 10) -> Response:
        """Sends a given request with internal session object.

        :param request: A request object that shall be sended
        :type request: requests.Request

        :param timeout: The time value for maximium waiting time of the response.
        :type int:

        :return: Returns the response of the given request
        :rtype: requests.Response

        """
        return self.session.send(request=request.prepare(), timeout=timeout)

    def get_capabilitites_request(self) -> OGCRequest:
        """Constructs a basic GetCapabilities request to use for requesting

        :return: A valid GetCapabilitites request
        :rtype: requests.Request
        """
        params = {
            "VERSION": self.capabilities.service_type.version,
            "REQUEST": "GetCapabilities",
            "SERVICE": self.capabilities.service_type.name
        }

        url = update_queryparams(
            url=self.capabilities.get_operation_url_by_name_and_method(
                "GetCapabilities", "Get").url,
            params=params)

        return OGCRequest(method="Get", url=url)



import datetime
from typing import Callable, Dict, List

from django.conf import settings
from django.contrib.gis.geos import Polygon
from django.utils import timezone
from isodate import (Duration, ISO8601Error, parse_date, parse_datetime,
                     parse_duration)
from isodate.isodatetime import datetime_isoformat
from isodate.isoduration import duration_isoformat

from ows_lib.xml_mapper.capabilities.mixins import OGCServiceMixin
from ows_lib.xml_mapper.mixins import CallbackList


class TimeExtent:
    """Helper class to abstract single time extent objects"""

    def __init__(
            self,
            start: datetime.datetime,
            stop: datetime.datetime = None,
            resolution: int = None,
            callback: Callable = None,
            *args,
            **kwargs):
        super().__init__(*args, **kwargs)
        self.__start = start
        self.__stop = stop
        self.__resolution = resolution
        self._callback = callback

    @property
    def start(self) -> datetime.datetime:
        return self.__start

    @start.setter
    def start(self, value: datetime.datetime) -> None:
        self.__start = value
        if self._callback:
            self._callback(self)

    @property
    def stop(self) -> datetime.datetime:
        return self.__stop

    @stop.setter
    def stop(self, value: datetime.datetime) -> None:
        self.__stop = value
        if self._callback:
            self._callback(self)

    @property
    def resolution(self) -> int:
        return self.__resolution

    @resolution.setter
    def resolution(self, value: int) -> None:
        self.__resolution = value
        if self._callback:
            self._callback(self)

    @property
    def is_value(self):
        return self.start and not self.stop and not self.resolution

    @property
    def is_interval(self):
        return self.start and self.stop and self.resolution

    @property
    def is_valid(self):
        return self.is_value or self.is_interval

    def to_xml_value(self) -> str:
        if self.is_value:
            return datetime_isoformat(self.start)
        elif self.is_interval:
            if self.resolution == 0:
                return f"{datetime_isoformat(self.start)}/{datetime_isoformat(self.stop)}/0"
            else:
                return f"{datetime_isoformat(self.start)}/{datetime_isoformat(self.stop)}/{duration_isoformat(self.resolution)}"

    def __str__(self) -> str:
        return f"{self.start} | {self.stop} | {self.resolution}"

    def transform_to_model(self) -> Dict:
        attr = {}
        if self.start:
            attr.update({"start": self.start})
        if self.stop:
            attr.update({"stop": self.stop})
        if self.resolution:
            attr.update({"resolution": self.resolution})
        return attr


class TimeDimensionMixin:
    # cache variable to store the parsed extent value
    _time_extents: CallbackList = CallbackList([], None)

    @property
    def time_extents(self):
        if not self._time_extents:
            self.__parse_extent()
        return self._time_extents

    @time_extents.setter
    def time_extents(self, time_extents: List[TimeExtent]) -> None:
        """Custom setter function to serialize a list of TimeExtent objects and update it to the xml node"""
        values = []
        intervals = []
        not_valid = []

        for time_extent in time_extents:
            if not time_extent._callback:
                time_extent._callback = self.__parse_extent
            if not time_extent.is_valid:
                raise ValueError(
                    f"time extent objects is not valid: {time_extent}")
            if time_extent.is_value:
                values.append(time_extent)
            elif time_extent.is_interval:
                intervals.append(time_extent)

        if values and intervals:
            raise ValueError("mixing values and intervals is not supported.")

        if not_valid:
            raise ValueError("some passed time extent objects are not valid")
        self._time_extents = CallbackList(
            time_extents, callback=self.__serialize_time_extents)
        self.__serialize_time_extents()

    def __serialize_time_extents(self, *args, **kwargs):
        self._extent = ",".join([time_extent.to_xml_value()
                                 for time_extent in self.time_extents])

    def __parse_extent_value(self, start, stop, resolution) -> tuple:
        _start = self.__parse_datetime_or_date(start)  # iso date time
        _stop = self.__parse_datetime_or_date(stop)  # iso date time
        try:
            _resolution = int(resolution)
        except ValueError:
            _resolution = parse_duration(resolution)
            if isinstance(_resolution, Duration):
                _resolution = _resolution.totimedelta(start=timezone.now())
        return _start, _stop, _resolution

    def __parse_datetime_or_date(self, value):
        _value = None
        try:
            _value = parse_datetime(value)
        except ISO8601Error:
            try:
                _value = parse_date(value)
            except ISO8601Error:
                settings.ROOT_LOGGER.debug(
                    msg=f"can't parse time dimension from value: {value}")

        return _value

    def __parse_list_of_multiple_intervals(self) -> List[TimeExtent]:
        __extents = []
        intervals = self._extent.split(",")
        for interval in intervals:
            __extents.append(self.__parse_single_interval(interval=interval))
        return __extents

    def __parse_single_interval(self, interval) -> TimeExtent:
        split = interval.split("/")
        start, stop, resolution = self.__parse_extent_value(
            start=split[0], stop=split[1], resolution=split[2])
        return TimeExtent(start=start, stop=stop, resolution=resolution)

    def __parse_list_of_values(self) -> List[TimeExtent]:
        split = self._extent.split(",")
        _extents = []
        for value in split:
            _value = self.__parse_datetime_or_date(value)
            if _value:
                _extents.append(
                    TimeExtent(start=_value, stop=_value))
        return _extents

    def __parse_extent(self, *args, **kwargs):
        """
            OGC WMS 1.3.0 Spech page 53:

            Table C.2 — Syntax for listing one or more extent values

            value 								A single value.
            value1,value2,value3,... 			a A list of multiple values.
            min/max/resolution 					An interval defined by its lower and upper bounds and its resolution.
            min1/max1/res1,min2/max2/res2,... 	a A list of multiple intervals.

            A resolution value of zero (as in min/max/0) means that the data are effectively at infinitely-fine resolution for the
            purposes of making requests on the server. For instance, an instrument which continuously monitors randomly-
            occurring data may have no explicitly defined temporal resolution.
        """

        # ogc wms dimension supports more thant time in iso 8601 format.
        # But for now we only implement this, cause other ones are not common usage
        __extents = []
        if self.name == "time" and self.units == "ISO8601":
            if "," in self._extent and "/" in self._extent:
                # case 4 of table C.2: A list of multiple interval
                __extents.extend(self.__parse_list_of_multiple_intervals())
            elif "/" in self._extent:
                # case 3 of table C.2: An interval defined by its lower and upper bounds and its resolution
                __extents.append(self.__parse_single_interval(
                    interval=self._extent))
            elif "," in self._extent:
                # case 2 of table C.2: a A list of multiple values
                __extents.extend(self.__parse_list_of_values())
            else:
                # case 1 of table C.2: one single value was detected
                _value = self.__parse_datetime_or_date(self._extent)
                if _value:
                    __extents.append(TimeExtent(start=_value, stop=_value))
        self._time_extents = CallbackList(
            __extents, callback=self.__serialize_time_extents)


class LayerMixin:
    def __str__(self):
        return self.identifier

    @property
    def _bbox_min_x(self):
        raise NotImplementedError

    @property
    def _bbox_min_y(self):
        raise NotImplementedError

    @property
    def _bbox_max_x(self):
        raise NotImplementedError

    @property
    def _bbox_max_y(self):
        raise NotImplementedError

    @property
    def bbox_lat_lon(self) -> Polygon:
        # there is no default xmlmap field which parses to a geos polygon. So we convert it here.
        if self._bbox_min_x and self._bbox_max_x and self._bbox_min_y and self._bbox_max_y:
            return Polygon(
                (
                    (self._bbox_min_x, self._bbox_min_y),
                    (self._bbox_min_x, self._bbox_max_y),
                    (self._bbox_max_x, self._bbox_max_y),
                    (self._bbox_max_x, self._bbox_min_y),
                    (self._bbox_min_x, self._bbox_min_y)
                )
            )

    @bbox_lat_lon.setter
    def bbox_lat_lon(self, polygon: Polygon) -> None:
        # Custom setter function to mapp the geos Polygon object back to the xml attributes
        self._bbox_min_x = polygon.extent[0]
        self._bbox_min_y = polygon.extent[1]
        self._bbox_max_x = polygon.extent[2]
        self._bbox_max_y = polygon.extent[3]

    @property
    def descendants(self, include_self=True):
        descendants = [self] if include_self else []
        for child in self.children:
            descendants.extend(child.descendants)
        return descendants

    def transform_to_model(self) -> Dict:
        attr = super().transform_to_model()
        if self.bbox_lat_lon:
            attr.update({"bbox_lat_lon": self.bbox_lat_lon})
        return attr


class WebMapServiceMixin(OGCServiceMixin):
    """Abstract class for WebMapService xml mappers,
    which implements functionality for global usage."""

    _possible_operations = ["GetCapabilities", "GetMap",
                            "GetFeatureInfo", "DescribeLayer", "GetLegendGraphic", "GetStyles"]

    @property
    def layers(self):
        """Flat list of all layers of this web map service"""
        __layers = []
        __layers.extend(self.root_layer.descendants)
        return __layers

    def get_layer_by_identifier(self, identifier: str):
        return next((layer for layer in self.layers if layer.identifier == identifier), None)



from typing import Callable, Dict

from django.conf import settings
from eulxml import xmlmap
from eulxml.xmlmap import XmlObject
from lxml.etree import XPathEvalError


class CallbackList(list):

    def __init__(self, iterable, callback: Callable = None, *args, **kwargs) -> None:
        super().__init__(iterable, *args, **kwargs)
        self.callback = callback

    def append(self, item) -> None:
        super().append(item)
        if self.callback:
            self.callback(list_operation="append", items=item)

    def extend(self, items) -> None:
        super().extend(items)
        if self.callback:
            self.callback(list_operation="extend", items=items)

    def pop(self, index):
        operation_url_to_pop = self[index]
        super().pop(index)
        if self.callback:
            self.callback(list_operation="pop", items=operation_url_to_pop)

    def clear(self) -> None:
        super().clear()
        if self.callback:
            self.callback(list_operation="clear")

    def insert(self, index, obj) -> None:
        super().insert(index, obj)
        if self.callback:
            self.callback(list_operation="insert", items=obj)

    def remove(self, value) -> None:
        super().remove(value)
        if self.callback:
            self.callback(list_operation="remove", items=value)


class CustomXmlObject(XmlObject):
    """ Custom Xml Object class which implements some generic functions to get the db model class and all relevant field content
        as dict.
    """

    def __get_xml_mapper_fields(self) -> Dict:
        """ Return a dict which contains the key, value pairs of the given field attribute name as key and the
            attribute value it self as value.

            Examples:
                If the following two classes are given:

                class Nested(DBModelConverterMixin, xmlmap.XmlObject):
                    ...

                class SomeXmlObject(DBModelConverterMixin, xmlmap.XmlObject):
                    name = xmlmap.StringField('name')
                    nested = xmlmap.NodeField('nested', Nested)
                    nested_list = xmlmap.NodeListField('nested', Nested)
                    _something = xmlmap.StringField('something')

                The SomeXmlObject().get_field_dict() function return {'name': 'Something'}

        Returns:
            field_dict (dict): the dict which contains all simple fields of the object it self.

        """
        field_dict = {}

        for key in filter(lambda key: not key.startswith('_'), self._fields.keys()):
            try:
                if not (isinstance(self._fields.get(key), xmlmap.NodeField) or
                        isinstance(self._fields.get(key), xmlmap.NodeListField) or
                        isinstance(self._fields.get(key), xmlmap.StringListField)):
                    if isinstance(self._fields.get(key), xmlmap.SimpleBooleanField) and getattr(self, key) is None or \
                       isinstance(self._fields.get(key), xmlmap.StringField) and getattr(self, key) is None or \
                       isinstance(self._fields.get(key), xmlmap.IntegerField) and getattr(self, key) is None:
                        # we don't append None values, cause if we construct a model with key=None and the db field
                        # don't allow Null values but has a default for Boolean the db will raise integrity
                        # errors.
                        continue
                    field_dict.update({key: getattr(self, key)})
            except XPathEvalError as e:
                settings.ROOT_LOGGER.error(
                    msg=f"error during parsing field: {key} in class {self.__class__.__name__}")
                settings.ROOT_LOGGER.exception(
                    e, stack_info=True, exc_info=True)
        return field_dict

    def transform_to_model(self) -> Dict:
        """
        Returns a dict of public objects fields and there values
        """
        field_dict = self.__get_xml_mapper_fields()
        for key, value in self.__dict__.items():
            if key.startswith('_') or key[0].isupper() or key != 'context' or value is None or key in self._fields.keys():
                break
            field_dict.update({key: value})
        return field_dict




from eulxml.xmlmap import NodeListField, StringField, XmlObject

from ows_lib.xml_mapper.iso_metadata.iso_metadata import MdMetadata
from ows_lib.xml_mapper.namespaces import (CSW_2_0_2_NAMESPACE, GMD_NAMESPACE,
                                           XSI_NAMESPACE)


class GetRecordsResponse(XmlObject):
    ROOT_NAME = "GetRecordByIdResponse"
    ROOT_NS = CSW_2_0_2_NAMESPACE
    ROOT_NAMESPACES = {
        "csw": CSW_2_0_2_NAMESPACE,
        "gmd": GMD_NAMESPACE,
        "xsi": XSI_NAMESPACE
    }

    XSD_SCHEMA = "http://schemas.opengis.net/csw/2.0.2/CSW-discovery.xsd"

    schema_location = StringField(
        xpath="./@xsi:schemaLocation"
    )

    gmd_records = NodeListField(
        xpath="./gmd:MD_Metadata", node_class=MdMetadata)

    def __init__(self, node=None, context=None, **kwargs):
        super().__init__(node, context, **kwargs)
        if not self.schema_location:
            self.schema_location = "http://www.opengis.net/cat/csw/2.0.2 http://schemas.opengis.net/csw/2.0.2/CSW-discovery.xsd"



class InitialError(Exception):
    """Raised when a ogc client can't be initialized by a given capabilities document"""
    pass


class MissingQueryParam(Exception):
    """ Base class for missing query parameter errors. """
    pass


class MissingRequestParam(MissingQueryParam):
    """Raised when the "REQUEST" query param is missed"""

    def __init__(self, msg="'REQUEST' query parameter was missed", *args, **kwargs):
        super().__init__(msg, *args, **kwargs)


class MissingVersionParam(MissingQueryParam):
    """Raised when the "VERSION" query param is missed"""

    def __init__(self, msg="'VERSION' query parameter was missed", *args, **kwargs):
        super().__init__(msg, *args, **kwargs)


class MissingServiceParam(MissingQueryParam):
    """Raised when the "SERVICE" query param is missed"""

    def __init__(self, msg="'SERVICE' query parameter was missed", *args, **kwargs):
        super().__init__(msg, *args, **kwargs)


class MissingBboxParam(MissingQueryParam):
    """Raised when the "BBOX" query param is missed"""

    def __init__(self, msg="'BBOX' query parameter was missed", *args, **kwargs):
        super().__init__(msg, *args, **kwargs)


class MissingCrsParam(MissingQueryParam):
    """Raised when the coordinate reference system (SRS, CRS, srsName) query param is missed"""

    def __init__(self, msg="'SRS', 'CRS' or 'srsName' query parameter was missed", *args, **kwargs):
        super().__init__(msg, *args, **kwargs)



from ows_lib.client.wms.mixins import WebMapServiceMixin


class WebMapService(WebMapServiceMixin):
    """Concrete WebMapService client for version 1.1.1"""

    crs_qp = "SRS"
    get_map_operation_name = "GetMap"
    get_feature_info_operation_name = "GetFeatureInfo"



from typing import Dict, List

from django.contrib.gis.geos import GEOSGeometry
from django.db.models.query_utils import Q
from django.http.request import HttpRequest as DjangoRequest
from eulxml.xmlmap import XmlObject, load_xmlobject_from_string
from lark.exceptions import LarkError
from lxml.etree import XMLSyntaxError
from pygeofilter.backends.django.evaluate import to_filter
from pygeofilter.parsers.ecql import parse as parse_ecql
from pygeofilter.parsers.fes.util import NodeParsingError
from requests import Request

from ows_lib.client.enums import OGCOperationEnum
from ows_lib.client.exceptions import MissingBboxParam, MissingServiceParam
from ows_lib.client.utils import (construct_polygon_from_bbox_query_param,
                                  get_requested_feature_types,
                                  get_requested_layers, get_requested_records)
from ows_lib.contrib.fes.parser import parse as parse_fes
from ows_lib.xml_mapper.exceptions import (
    InvalidParameterValueException,
    MissingConstraintLanguageParameterException)
from ows_lib.xml_mapper.xml_requests.csw.get_records import GetRecordsRequest
from ows_lib.xml_mapper.xml_requests.csw.get_records_by_id import \
    GetRecordByIdRequest
from ows_lib.xml_mapper.xml_requests.utils import PostRequest
from ows_lib.xml_mapper.xml_requests.wfs.get_feature import (GetFeatureRequest,
                                                             Query)


class OGCRequest(Request):
    """Extended Request class which provides some analyzing functionality for ogc requests."""

    def __init__(self, django_request: DjangoRequest = None, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._djano_request = django_request
        self._ogc_query_params: Dict = {}
        self._bbox: GEOSGeometry = None
        self._requested_entities: List[str] = []
        self._xml_request: XmlObject = None
        self.operation = "unknown"
        self.service_version = "unknown"
        self.service_type = "unknown"

        if self.method == "GET":
            self.operation: str = self.ogc_query_params.get("REQUEST", "")
            self.service_version: str = self.ogc_query_params.get(
                "VERSION", "")
            self.service_type: str = self.ogc_query_params.get("SERVICE", "")
        elif self.method == "POST" and self.data and isinstance(self.data, bytes):
            post_request: PostRequest = load_xmlobject_from_string(
                string=self.data, xmlclass=PostRequest)

            self.operation: str = post_request.operation
            self.service_version: str = post_request.version
            self.service_type: str = post_request.service_type

    @classmethod
    def from_django_request(cls, request: DjangoRequest):
        """helper function to construct an OGCRequest from a django request object."""
        ogc_request = cls(
            method=request.method,
            url=request.build_absolute_uri(),
            params={**request.GET, **request.POST},
            data=request.body,
            cookies=request.COOKIES,
            django_request=request
        )
        return ogc_request

    @property
    def requested_entities(self) -> List[str]:
        """Returns the list of requested entities

        This function analyzes the request and find out which layers or featuretypes, or records are requested.

        :return: list of requested layers | list of request featuretypes
        :rtype: List[str]
        """
        if not self._requested_entities:
            if self.is_wms:
                self._requested_entities.extend(
                    get_requested_layers(params=self.ogc_query_params))
            elif self.is_get_feature_request:
                if self.is_get:
                    self._requested_entities.extend(
                        get_requested_feature_types(params=self.ogc_query_params))
                elif self.is_post:
                    self._requested_entities.extend(
                        self.xml_request.requested_feature_types)
            elif self.is_get_record_by_id_request:
                if self.is_get:
                    self._requested_entities.extend(
                        get_requested_records(params=self.ogc_query_params)
                    )
                elif self.is_post:
                    self._requested_entities.extend(
                        self.xml_request.ids)
            else:
                pass
        return self._requested_entities

    def filter_constraint(self, field_mapping=None, mapping_choices=None) -> Q:
        if self.is_csw and self.is_get_records_request:
            if self.is_get:
                constraint = self.ogc_query_params.get("Constraint", "")
                constraint_language = self.ogc_query_params.get(
                    "CONSTRAINTLANGUAGE", "")
                if constraint and not constraint_language:
                    return MissingConstraintLanguageParameterException(ogc_request=self)

                elif constraint and constraint_language:
                    if constraint_language == "CQL_TEXT":
                        try:
                            ast = parse_ecql(constraint)
                            return to_filter(ast, field_mapping, mapping_choices)
                        except LarkError as e:
                            return InvalidParameterValueException(ogc_request=self, message=e)
                    elif constraint_language == "FILTER":
                        try:
                            ast = parse_fes(constraint)
                            return to_filter(ast, field_mapping, mapping_choices)
                        except (XMLSyntaxError, NodeParsingError) as e:
                            return InvalidParameterValueException(
                                ogc_request=self,
                                message=e
                            )
                    else:
                        return InvalidParameterValueException(
                            ogc_request=self,
                            message="Provided CONSTRAINTLANGUAGE is not supported.")
            elif self.is_post:
                return self.xml_request.get_django_filter(field_mapping, mapping_choices)

        return Q()

    @property
    def is_wms(self) -> bool:
        """Check for wms request

        :return: true if this is a wms request
        :rtype: bool
        """
        return self.service_type.lower() == 'wms'

    @property
    def is_wfs(self) -> bool:
        """Check for wfs request

        :return: true if this is a wfs request
        :rtype: bool
        """
        return self.service_type.lower() == 'wfs'

    @property
    def is_csw(self) -> bool:
        """Check for wfs request

        :return: true if this is a wfs request
        :rtype: bool
        """
        return self.service_type.lower() == 'csw'

    @property
    def is_post(self) -> bool:
        """Check for post method

        :return: true if this is a post request
        :rtype: bool
        """
        return self.method.lower() == "post"

    @property
    def is_get(self) -> bool:
        """Check for get method

        :return: true if this is a post request
        :rtype: bool
        """
        return self.method.lower() == "get"

    @property
    def is_get_capabilities_request(self) -> bool:
        """Check for ogc get capabilites request

        :return: true if this is a get capabilities request
        :rtype: bool
        """
        return self.operation == OGCOperationEnum.GET_CAPABILITIES.value

    @property
    def is_get_map_request(self) -> bool:
        """Check for wms get map request

        :return: true if this is a wms get map request
        :rtype: bool
        """
        return self.operation == OGCOperationEnum.GET_MAP.value

    @property
    def is_get_feature_info_request(self) -> bool:
        """Check for wms transaction request

        :return: true if this is a wfs tranasction request
        :rtype: bool
        """
        return self.operation == OGCOperationEnum.GET_FEATURE_INFO.value

    @property
    def is_get_feature_request(self) -> bool:
        """Check for wfs get feature request

        :return: true if this is a wfs get feature request
        :rtype: bool
        """
        return self.operation == OGCOperationEnum.GET_FEATURE.value

    @property
    def is_transaction_request(self) -> bool:
        """Check for wfs transaction request

        :return: true if this is a wfs tranasction request
        :rtype: bool
        """
        return self.operation == OGCOperationEnum.TRANSACTION.value

    @property
    def is_describe_record_request(self) -> bool:
        return self.operation == OGCOperationEnum.DESCRIBE_RECORD.value

    @property
    def is_get_record_by_id_request(self) -> bool:
        return self.operation == OGCOperationEnum.GET_RECORD_BY_ID.value

    @property
    def is_get_records_request(self) -> bool:
        return self.operation == OGCOperationEnum.GET_RECORDS.value

    @property
    def bbox(self) -> GEOSGeometry:
        """Analyzes the given request and tries to construct a Polygon from the query parameters.

        The axis order for different wms/wfs versions will be well transformed to the correct needed mathematical interpretation.

        :raises MissingBboxParam: if the bbox query param is missing
        :raises MissingServiceParam: if the service query param is missing. Without that the correct axis order can't be interpreted.

        :return: the given bbox as polygon object
        :rtype: GEOSGeometry
        """
        if not self._bbox:
            try:
                self._bbox = construct_polygon_from_bbox_query_param(
                    get_dict=self.ogc_query_params)
            except (MissingBboxParam, MissingServiceParam):
                # only to avoid error while handling sql in service property
                self._bbox = GEOSGeometry("POLYGON EMPTY")
        return self._bbox

    @property
    def ogc_query_params(self) -> Dict:
        """ Parses the GET parameters into all member variables, which can be found in a ogc request.

        :return: all ogc query parameters
        :rtype: Dict
        """
        if not self._ogc_query_params:
            query_keys = ["SERVICE", "REQUEST", "LAYERS", "BBOX", "VERSION", "FORMAT",
                          "OUTPUTFORMAT", "SRS", "CRS", "SRSNAME", "WIDTH", "HEIGHT",
                          "TRANSPARENT", "EXCEPTIONS", "BGCOLOR", "TIME", "ELEVATION",
                          "QUERY_LAYERS", "INFO_FORMAT", "FEATURE_COUNT", "I", "J",
                          "NAMESPACE", "resultType", "requestId", "TypeName", "outputFormat",
                          "outputSchema", "startPosition", "maxRecords", "schemaLanguage",
                          "ElementSetName", "ElementName", "typeNames", "CONSTRAINTLANGUAGE",
                          "Constraint", "SortBy", "DistributedSearch",
                          "hopCount", "ResponseHandler", "Id"]

            for key in query_keys:
                value = self.params.get(key, self.params.get(key.lower(), ""))

                if value:
                    if isinstance(value, list):
                        # if multiple values are passed in multiple queryparams, we pick the first item
                        value = value[0]

                    self._ogc_query_params.update({
                        key: value
                    })

        return self._ogc_query_params

    @property
    def xml_request(self) -> XmlObject:
        """Constructs a xml request object based on the given request.

        This function analyzes the given request by its method and operation. 
        If it is a get request, the get feature operation for example will be converted to an postable xml object.

        :return: The mapped xml object
        :rtype: XmlObject
        """
        if not self._xml_request:
            # TODO: implement the xml request generation for other requests too.
            if self.is_get_feature_request:  # NOSONAR: See todo above
                # FIXME: depending on version, different xml mapper are needed...
                if self.is_post:
                    self._xml_request: GetFeatureRequest = load_xmlobject_from_string(
                        string=self.data, xmlclass=GetFeatureRequest)
                elif self.is_get:
                    # we construct a xml get feature request to post it with a filter
                    queries: List[Query] = []
                    for feature_type in self.requested_entities:
                        query: Query = Query()
                        query.type_names = feature_type
                        queries.append(query)
                    self._xml_request: GetFeatureRequest = GetFeatureRequest()
                    self._xml_request.queries = queries
            elif self.is_get_records_request:
                if self.is_post:
                    self._xml_request: GetRecordsRequest = load_xmlobject_from_string(
                        string=self.data, xmlclass=GetRecordsRequest)
                elif self.is_get:
                    sort_by = self.ogc_query_params.get("SortBy")
                    if sort_by:
                        element_name, a_or_d = sort_by.split(":")
                        if a_or_d == "A":
                            sort_by = element_name
                        elif a_or_d == "D":
                            sort_by = f"-{element_name}"
                    constraint = self.ogc_query_params.get("Constraint")
                    constraint_language = self.ogc_query_params.get(
                        "CONSTRAINTLANGUAGE")
                    self._xml_request = GetRecordsRequest()
                    self._xml_request.service_version = self.ogc_query_params.get(
                        "VERSION")
                    self._xml_request.service_type = self.ogc_query_params.get(
                        "SERVICE")
                    self._xml_request.result_type = self.ogc_query_params.get(
                        "resultType")
                    self._xml_request.start_position = self.ogc_query_params.get(
                        "startPosition")
                    self._xml_request.max_records = self.ogc_query_params.get(
                        "maxRecords")
                    self._xml_request.element_set_name = self.ogc_query_params.get(
                        "ElementSetName")
                    self._xml_request.sort_by = sort_by
                    if constraint and constraint_language == "CQL_TEXT":
                        self._xml_request.cql_filter = constraint
                        self._xml_request.constraint_version = "1.1.0"
                    elif constraint and constraint_language == "FILTER":
                        self._xml_request.fes_filter = load_xmlobject_from_string(
                            constraint, XmlObject)
                        self._xml_request.constraint_version = "1.1.0"
            elif self.is_get_record_by_id_request:
                if self.is_post:
                    self._xml_request: GetRecordByIdRequest = load_xmlobject_from_string(
                        string=self.data, xmlclass=GetRecordByIdRequest)
                elif self.is_get:
                    self._xml_request = GetRecordByIdRequest()
                    self._xml_request.service_version = self.ogc_query_params.get(
                        "VERSION")
                    self._xml_request.service_type = self.ogc_query_params.get(
                        "SERVICE")
                    self._xml_request.element_set_name = self.ogc_query_params.get(
                        "ElementSetName")
                    self._xml_request.ids = self.ogc_query_params.get(
                        "id", "").split(",")
        return self._xml_request



from enum import Enum


class OGCServiceEnum(Enum):
    """ Defines all supported service types

    """
    ALL = "all"
    WMS = "wms"
    WFS = "wfs"
    WMC = "wmc"
    DATASET = "dataset"
    CSW = "csw"


class OGCOperationEnum(Enum):
    """ Defines all known operation names

    """
    # ALL
    GET_CAPABILITIES = "GetCapabilities"

    # WMS
    GET_MAP = "GetMap"
    GET_FEATURE_INFO = "GetFeatureInfo"
    DESCRIBE_LAYER = "DescribeLayer"
    GET_LEGEND_GRAPHIC = "GetLegendGraphic"
    GET_STYLES = "GetStyles"
    PUT_STYLES = "PutStyles"

    # WFS
    GET_FEATURE = "GetFeature"
    TRANSACTION = "Transaction"
    LOCK_FEATURE = "LockFeature"
    DESCRIBE_FEATURE_TYPE = "DescribeFeatureType"
    GET_FEATURE_WITH_LOCK = "GetFeatureWithLock"
    GET_GML_OBJECT = "GetGmlObject"
    LIST_STORED_QUERIES = "ListStoredQueries"
    GET_PROPERTY_VALUE = "GetPropertyValue"
    DESCRIBE_STORED_QUERIES = "DescribeStoredQueries"

    # CSW
    GET_RECORDS = "GetRecords"
    DESCRIBE_RECORD = "DescribeRecord"
    GET_DOMAIN = "GetDomain"
    GET_RECORD_BY_ID = "GetRecordById"
    HARVEST_RECORDS = "harvestRecords"



from eulxml.xmlmap import (FloatField, IntegerField, NodeField, NodeListField,
                           SimpleBooleanField, StringField, StringListField)

from ows_lib.xml_mapper.capabilities.mixins import (OGCServiceTypeMixin,
                                                    ReferenceSystemMixin)
from ows_lib.xml_mapper.capabilities.wms.mixins import (LayerMixin,
                                                        TimeDimensionMixin,
                                                        WebMapServiceMixin)
from ows_lib.xml_mapper.mixins import CustomXmlObject
from ows_lib.xml_mapper.namespaces import XLINK_NAMESPACE


class WebMapServiceDefaultSettings(CustomXmlObject):
    ROOT_NAMESPACES = {
        "xlink": XLINK_NAMESPACE
    }


class ServiceMetadataContact(WebMapServiceDefaultSettings):
    ROOT_NAME = "ContactInformation"

    name = StringField(
        xpath="ContactPersonPrimary/ContactOrganization")
    person_name = StringField(
        xpath="ContactPersonPrimary/ContactPerson")
    phone = StringField(xpath="ContactVoiceTelephone")
    facsimile = StringField(xpath="ContactFacsimileTelephone")
    email = StringField(xpath="ContactElectronicMailAddress")
    country = StringField(xpath="ContactAddress/Country")
    postal_code = StringField(xpath="ContactAddress/PostCode")
    city = StringField(xpath="ContactAddress/City")
    state_or_province = StringField(
        xpath="ContactAddress/StateOrProvince")
    address = StringField(xpath="ContactAddress/Address")


class ServiceType(OGCServiceTypeMixin, WebMapServiceDefaultSettings):
    ROOT_NAME = "WMS_Capabilities/@version='1.3.0'"

    version = StringField(xpath="./@version", choices='1.3.0')

    _name = StringField(xpath="./Service/Name")


class TimeDimension(TimeDimensionMixin, WebMapServiceDefaultSettings):
    """ Time Dimension in ISO8601 format"""

    ROOT_NAME = "Dimension[@name='time']"

    name = StringField(xpath="./@name", choices="time")
    units = StringField(xpath="./@units", choices="ISO8601")

    _extent = StringField(xpath="../Extent[@name='time']")


class ReferenceSystem(ReferenceSystemMixin, WebMapServiceDefaultSettings):
    ROOT_NAME = "SRS"

    _ref_system = StringField(xpath=".")


class MimeType(WebMapServiceDefaultSettings):
    ROOT_NAME = "Format"
    mime_type = StringField(xpath="./Format")


class LegendUrl(WebMapServiceDefaultSettings):
    ROOT_NAME = "LegendUrl"

    legend_url = StringField(
        xpath="./OnlineResource/@xlink:href")
    height = IntegerField(xpath="./@height")
    width = IntegerField(xpath="./@width")
    mime_type = NodeField(xpath="./Format", node_class=MimeType)


class Style(WebMapServiceDefaultSettings):
    ROOT_NAME = "Style"

    name = StringField(xpath="./Name")
    title = StringField(xpath="./Title")
    legend_url = NodeField(xpath="./LegendURL", node_class=LegendUrl)


class RemoteMetadata(WebMapServiceDefaultSettings):
    ROOT_NAME = "OnlineResource/@xlink:href"
    link = StringField(
        xpath="./@xlink:href")


class Layer(LayerMixin, WebMapServiceDefaultSettings):
    ROOT_NAME = "Layer"

    title = StringField(xpath="./Title")
    abstract = StringField(xpath="./Abstract")
    keywords = StringListField(xpath="./KeywordList/Keyword")

    scale_min = FloatField(xpath="./ScaleHint/@min")
    scale_max = FloatField(xpath="./ScaleHint/@max")

    _bbox_min_x = FloatField(
        xpath="./LatLonBoundingBox/@minx")
    _bbox_max_x = FloatField(
        xpath="./LatLonBoundingBox/@maxx")
    _bbox_min_y = FloatField(
        xpath="./LatLonBoundingBox/@miny")
    _bbox_max_y = FloatField(
        xpath="./LatLonBoundingBox/@maxy")

    reference_systems = NodeListField(
        xpath="./SRS", node_class=ReferenceSystem)
    identifier = StringField(xpath="./Name")
    styles = NodeListField(xpath="./Style", node_class=Style)
    is_queryable = SimpleBooleanField(
        xpath="./@queryable", true=1, false=0)
    is_opaque = SimpleBooleanField(xpath="./@opaque", true=1, false=0)
    is_cascaded = SimpleBooleanField(
        xpath="./@cascaded", true=1, false=0)

    dimensions = NodeListField(
        xpath="./Dimension[@name='time']", node_class=TimeDimension)

    parent = NodeField(xpath="../../Layer", node_class="self")
    children = NodeListField(xpath="./Layer", node_class="self")

    remote_metadata = NodeListField(
        xpath="./MetadataURL/OnlineResource",
        node_class=RemoteMetadata)


class WebMapService(WebMapServiceMixin, WebMapServiceDefaultSettings):
    ROOT_NAME = "WMT_MS_Capabilities/@version='1.1.1'"
    XSD_SCHEMA = "http://schemas.opengis.net/wms/1.1.1/WMS_MS_Capabilities.dtd"

    title = StringField(xpath="./Service/Title")
    abstract = StringField(xpath="./Service/Abstract")
    fees = StringField(xpath="./Service/Fees")
    access_constraints = StringField(xpath="./Service/AccessConstraints")

    # ForeignKey
    service_contact = NodeField(xpath="./Service/ContactInformation",
                                node_class=ServiceMetadataContact)

    # ManyToManyField
    keywords = StringListField(xpath="./Service/KeywordList/Keyword")

    service_url = StringField(
        xpath="./Service/OnlineResource/@xlink:href")

    service_type = NodeField(xpath=".", node_class=ServiceType)

    root_layer = NodeField(
        xpath="./Capability/Layer", node_class=Layer)

    _layers = NodeListField(
        xpath="./Capability//Layer", node_class=Layer)

    # cause the information of operation urls are stored as entity name inside the xpath, we need to parse every operation url seperate.
    # To simplify the access of operation_urls we write a custom getter and setter property for it.
    # With that technique the usage of this mapper is easier and matches the db model
    _get_capabilities_mime_types = StringListField(
        xpath="./Capability/Request/GetCapabilities/Format")
    _get_capabilities_get_url = StringField(
        xpath="./Capability/Request/GetCapabilities/DCPType/HTTP/Get/OnlineResource/@xlink:href")
    _get_capabilities_post_url = StringField(
        xpath="./Capability/Request/GetCapabilities/DCPType/HTTP/Post/OnlineResource/@xlink:href")

    _get_map_mime_types = StringListField(
        xpath="./Capability/Request/GetMap/Format")
    _get_map_get_url = StringField(
        xpath="./Capability/Request/GetMap/DCPType/HTTP/Get/OnlineResource/@xlink:href")
    _get_map_post_url = StringField(
        xpath="./Capability/Request/GetMap/DCPType/HTTP/Post/OnlineResource/@xlink:href")

    _get_feature_info_mime_types = StringListField(
        xpath="./Capability/Request/GetFeatureInfo/Format")
    _get_feature_info_get_url = StringField(
        xpath="./Capability/Request/GetFeatureInfo/DCPType/HTTP/Get/OnlineResource/@xlink:href")
    _get_feature_info_post_url = StringField(
        xpath="./Capability/Request/GetFeatureInfo/DCPType/HTTP/Post/OnlineResource/@xlink:href")

    _describe_layer_mime_types = StringListField(
        xpath="./Capability/Request/DescribeLayer/Format")
    _describe_layer_get_url = StringField(
        xpath="./Capability/Request/DescribeLayer/DCPType/HTTP/Get/OnlineResource/@xlink:href")
    _describe_layer_post_url = StringField(
        xpath="./Capability/Request/DescribeLayer/DCPType/HTTP/Post/OnlineResource/@xlink:href")

    _get_legend_graphic_mime_types = StringListField(
        xpath="./Capability/Request/GetLegendGraphic/Format")
    _get_legend_graphic_get_url = StringField(
        xpath="./Capability/Request/GetLegendGraphic/DCPType/HTTP/Get/OnlineResource/@xlink:href")
    _get_legend_graphic_post_url = StringField(
        xpath="./Capability/Request/GetLegendGraphic/DCPType/HTTP/Post/OnlineResource/@xlink:href")

    _get_styles_mime_types = StringListField(
        xpath="./Capability/Request/GetStyles/Format")
    _get_styles_get_url = StringField(
        xpath="./Capability/Request/GetStyles/DCPType/HTTP/Get/OnlineResource/@xlink:href")
    _get_styles_post_url = StringField(
        xpath="./Capability/Request/GetStyles/DCPType/HTTP/Post/OnlineResource/@xlink:href")



from eulxml.xmlmap import NodeField, StringField, StringListField

from ows_lib.xml_mapper.capabilities.mixins import (OGCServiceMixin,
                                                    OGCServiceTypeMixin)
from ows_lib.xml_mapper.mixins import CustomXmlObject
from ows_lib.xml_mapper.namespaces import (CSW_2_0_2_NAMESPACE, OWS_NAMESPACE,
                                           XLINK_NAMESPACE)


class CatalogueServiceDefaultSettings(CustomXmlObject):
    ROOT_NS = CSW_2_0_2_NAMESPACE
    ROOT_NAMESPACES = {
        "csw": CSW_2_0_2_NAMESPACE,
        "ows": OWS_NAMESPACE,
        "xlink": XLINK_NAMESPACE
    }


class ServiceMetadataContact(CatalogueServiceDefaultSettings):
    ROOT_NAME = "ServiceProvider"
    ROOT_NS = OWS_NAMESPACE

    name = StringField(
        xpath="./ows:ProviderName")

    person_name = StringField(
        xpath="./ows:ServiceContact/ows:IndividualName")

    phone = StringField(
        xpath="./ows:ServiceContact/ows:ContactInfo/ows:Phone/ows:Voice")
    facsimile = StringField(
        xpath="./ows:ServiceContact/ows:ContactInfo/ows:Phone/ows:Facsimile")
    email = StringField(
        xpath="./ows:ServiceContact/ows:ContactInfo/ows:Address/ows:ElectronicMailAddress")
    country = StringField(
        xpath="./ows:ServiceContact/ows:ContactInfo/ows:Address/ows:Country")
    postal_code = StringField(
        xpath="./ows:ServiceContact/ows:ContactInfo/ows:Address/ows:PostalCode")
    city = StringField(
        xpath="./ows:ServiceContact/ows:ContactInfo/ows:Address/ows:City")
    state_or_province = StringField(
        xpath="./ows:ServiceContact/ows:ContactInfo/ows:Address/ows:AdministrativeArea")
    address = StringField(
        xpath="./ows:ServiceContact/ows:ContactInfo/ows:Address/ows:DeliveryPoint")


class ServiceType(OGCServiceTypeMixin, CatalogueServiceDefaultSettings):
    ROOT_NAME = "Capabilities"

    version = StringField(xpath="./@version", choices='2.0.2')
    _name = StringField(xpath="./ows:ServiceIdentification/ows:ServiceType")


class CatalogueService(OGCServiceMixin, CatalogueServiceDefaultSettings):
    ROOT_NAME = "Capabilities"
    XSD_SCHEMA = "http://schemas.opengis.net/csw/2.0.2/CSW-discovery.xsd"

    _possible_operations = ["GetCapabilities", "DescribeRecord",
                            "GetDomain", "GetRecords", "GetRecordById"]

    service_type = NodeField(xpath=".", node_class=ServiceType)
    title = StringField(xpath="./ows:ServiceIdentification/ows:Title")
    abstract = StringField(xpath="./ows:ServiceIdentification/ows:Abstract")
    fees = StringField(xpath="./ows:ServiceIdentification/ows:Fees")
    access_constraints = StringField(
        xpath="./ows:ServiceIdentification/ows:AccessConstraints")

    # ForeignKey
    service_contact = NodeField(xpath="./ows:ServiceProvider",
                                node_class=ServiceMetadataContact)

    # ManyToManyField
    keywords = StringListField(
        xpath="./ows:ServiceIdentification/ows:Keywords/ows:Keyword")

    # cause the information of operation urls are stored as entity name inside the xpath, we need to parse every operation url seperate.
    # To simplify the access of operation_urls we write a custom getter and setter property for it.
    # With that technique the usage of this mapper is easier and matches the db model
    _get_capabilities_mime_types = StringListField(
        xpath="./ows:OperationsMetadata/ows:Operation[@name='GetCapabilities']/ows:Parameter[@name='AcceptFormats']/ows:AllowedValues/ows:Value")
    _get_capabilities_get_url = StringField(
        xpath="./ows:OperationsMetadata/ows:Operation[@name='GetCapabilities']/ows:DCP/ows:HTTP/ows:Get/@xlink:href")
    _get_capabilities_post_url = StringField(
        xpath="./ows:OperationsMetadata/ows:Operation[@name='GetCapabilities']/ows:DCP/ows:HTTP/ows:Post/@xlink:href")
    _describe_record_mime_types = StringListField(
        xpath="./ows:OperationsMetadata/ows:Operation[@name='DescribeRecord']/ows:Parameter[@name='outputFormat']/ows:Value")
    _describe_record_get_url = StringField(
        xpath="./ows:OperationsMetadata/ows:Operation[@name='DescribeRecord']/ows:DCP/ows:HTTP/ows:Get/@xlink:href")
    _describe_record_post_url = StringField(
        xpath="./ows:OperationsMetadata/ows:Operation[@name='DescribeRecord']/ows:DCP/ows:HTTP/ows:Post/@xlink:href")
    _get_domain_mime_types = StringListField(
        xpath="./ows:OperationsMetadata/ows:Operation[@name='GetDomain']/ows:Parameter[@name='ParameterName']/ows:Value")
    _get_domain_get_url = StringField(
        xpath="./ows:OperationsMetadata/ows:Operation[@name='GetDomain']/ows:DCP/ows:HTTP/ows:Get/@xlink:href")
    _get_domain_post_url = StringField(
        xpath="./ows:OperationsMetadata/ows:Operation[@name='GetDomain']/ows:DCP/ows:HTTP/ows:Post/@xlink:href")
    _get_records_mime_types = StringListField(
        xpath="./ows:OperationsMetadata/ows:Operation[@name='GetRecords']/ows:Parameter[@name='outputFormat']/ows:Value")
    _get_records_get_url = StringField(
        xpath="./ows:OperationsMetadata/ows:Operation[@name='GetRecords']/ows:DCP/ows:HTTP/ows:Get/@xlink:href")
    _get_records_post_url = StringField(
        xpath="./ows:OperationsMetadata/ows:Operation[@name='GetRecords']/ows:DCP/ows:HTTP/ows:Post/@xlink:href")
    _get_record_by_id_mime_types = StringListField(
        xpath="./ows:OperationsMetadata/ows:Operation[@name='GetRecordById']/ows:Parameter[@name='outputFormat']/ows:Value")
    _get_record_by_id_get_url = StringField(
        xpath="./ows:OperationsMetadata/ows:Operation[@name='GetRecordById']/ows:DCP/ows:HTTP/ows:Get/@xlink:href")
    _get_record_by_id_post_url = StringField(
        xpath="./ows:OperationsMetadata/ows:Operation[@name='GetRecordById']/ows:DCP/ows:HTTP/ows:Post/@xlink:href")

    get_records_constraints = StringListField(
        xpath="./ows:OperationsMetadata/ows:Operation[@name='GetRecords']/ows:Constraint/ows:Value")
